from dataclasses import dataclass
import importlib
from pathlib import Path
from typing import Callable

from appdirs import user_cache_dir

from mlebench.grade_helpers import Grader
from mlebench.utils import get_logger, get_module_dir, get_repo_dir, import_fn, load_yaml

logger = get_logger(__name__)


DEFAULT_DATA_DIR = (Path(user_cache_dir()) / "mle-bench" / "data").resolve()


@dataclass(frozen=True)
class Competition:
    id: str
    name: str
    description: str
    grader: Grader
    answers: Path
    gold_submission: Path
    sample_submission: Path
    competition_type: str
    prepare_fn: Callable[[Path, Path, Path], Path]
    raw_dir: Path
    private_dir: Path
    public_dir: Path
    checksums: Path
    leaderboard: Path

    def __post_init__(self):
        assert isinstance(self.id, str), "Competition id must be a string."
        assert isinstance(self.name, str), "Competition name must be a string."
        assert isinstance(self.description, str), "Competition description must be a string."
        assert isinstance(self.grader, Grader), "Competition grader must be of type Grader."
        assert isinstance(self.answers, Path), "Competition answers must be a Path."
        assert isinstance(self.gold_submission, Path), "Gold submission must be a Path."
        assert isinstance(self.sample_submission, Path), "Sample submission must be a Path."
        assert isinstance(self.competition_type, str), "Competition type must be a string."
        assert isinstance(self.checksums, Path), "Checksums must be a Path."
        assert isinstance(self.leaderboard, Path), "Leaderboard must be a Path."
        assert len(self.id) > 0, "Competition id cannot be empty."
        assert len(self.name) > 0, "Competition name cannot be empty."
        assert len(self.description) > 0, "Competition description cannot be empty."
        assert len(self.competition_type) > 0, "Competition type cannot be empty."

    @staticmethod
    def from_dict(data: dict) -> "Competition":
        grader = Grader.from_dict(data["grader"])

        try:
            return Competition(
                id=data["id"],
                name=data["name"],
                description=data["description"],
                grader=grader,
                answers=data["answers"],
                sample_submission=data["sample_submission"],
                gold_submission=data["gold_submission"],
                competition_type=data["competition_type"],
                prepare_fn=data["prepare_fn"],
                raw_dir=data["raw_dir"],
                public_dir=data["public_dir"],
                private_dir=data["private_dir"],
                checksums=data["checksums"],
                leaderboard=data["leaderboard"],
            )
        except KeyError as e:
            raise ValueError(f"Missing key {e} in competition config!")


class Registry:
    def __init__(self, data_dir: Path = DEFAULT_DATA_DIR, registry_dir: Path = None):
        self._data_dir = data_dir.resolve()
        self._custom_registry_dir = registry_dir.resolve() if registry_dir else None
        self.mode = 'test'

    def _coerce_file_import(
        self, fn_import_string: str, root_dir: Path, competition_id: str
    ) -> str:
        if fn_import_string.startswith("file:"):
            return fn_import_string

        module_name, fn_name = fn_import_string.split(":")
        try:
            importlib.import_module(module_name)
            return fn_import_string
        except ModuleNotFoundError as exc:
            if exc.name != module_name and not module_name.startswith(f"{exc.name}."):
                raise

            leaf = module_name.split(".")[-1]
            file_module = root_dir / competition_id / f"{leaf}.py"
            if not file_module.exists() and leaf.endswith("_val"):
                fallback = root_dir / competition_id / f"{leaf[:-4]}.py"
                if fallback.exists():
                    file_module = fallback

            if not file_module.exists():
                raise

            return f"file:{file_module}:{fn_name}"

    def _resolve_competition_root(self, competition_id: str) -> Path:
        """
        Resolve where a competition config lives.
        - Prefer custom registry_dir if provided by user.
        - Prefer top-level `dabench/` for DABench-prefixed tasks.
        - Check `data_dir` for user-uploaded tasks.
        - Fallback to legacy `mlebench/competitions/` for everything else.
        """
        # Priority 1: Use custom registry_dir if provided
        if self._custom_registry_dir:
            if (self._custom_registry_dir / competition_id / "config.yaml").exists():
                return self._custom_registry_dir
            # If custom registry_dir doesn't have config, still try to use it
            # (user might have custom structure)
            return self._custom_registry_dir

        # Priority 2: DABench tasks
        repo_dir = get_repo_dir()
        dabench_root = repo_dir / "benchmarks" / "dabench" / "competitions"
        if competition_id.startswith("dabench-") and (dabench_root / competition_id).exists():
            return dabench_root

        # Priority 3: Check if the competition is in the data directory (user uploaded)
        if (self._data_dir / competition_id / "config.yaml").exists():
            return self._data_dir

        if (legacy_root / competition_id).exists():
            return legacy_root
        if (dabench_root / competition_id).exists():
            return dabench_root
        return legacy_root

    def set_mode(self, mode: str = 'test'):
        """Set the mode of the registry.
        Args:
            mode: The mode of the registry. Can be 'test' or 'validation'.
        """
        assert mode in ['test', 'validation', 'prepare'], "Mode must be in ['test', 'validation', 'prepare']."
        self.mode = mode

    def get_competition(self, competition_id: str) -> Competition:
        """Fetch the competition from the registry."""

        root_dir = self._resolve_competition_root(competition_id)
        is_dabench = root_dir.name == "competitions" and root_dir.parent.name == "dabench"

        config_path = root_dir / competition_id / "config.yaml"
        config = load_yaml(config_path)

        checksums_path = root_dir / competition_id / "checksums.yaml"
        leaderboard_path = root_dir / competition_id / "leaderboard.csv"

        # Resolve description file. DABench configs may still point to legacy paths.
        if root_dir.name == "competitions" and root_dir.parent.name == "dabench":
            description_path = root_dir / competition_id / "description.md"
        else:
            # Try to find description relative to competition dir first
            candidate_desc = root_dir / competition_id / config["description"]
            if candidate_desc.exists():
                description_path = candidate_desc
            else:
                description_path = get_repo_dir() / config["description"]
                if not description_path.exists() and config["description"].startswith("mlebench/"):
                    description_path = get_repo_dir() / "benchmarks" / config["description"]
        description = description_path.read_text()

        # Config for different modes
        base_preparer = config["preparer"]
        base_answers = config["dataset"]["answers"]
        base_sample_submission = config["dataset"]["sample_submission"]

        config_preparer = base_preparer
        config_answers = base_answers
        config_sample_submission = base_sample_submission
        public_folder = 'public'
        private_folder = 'private'

        if is_dabench:
            # DABench evaluation should NOT depend on any prepare logic; always use the
            # existing prepared/public + prepared/private folders.
            config_preparer = base_preparer
            config_answers = base_answers
            config_sample_submission = base_sample_submission
            public_folder = "public"
            private_folder = "private"
        else:
            if self.mode == 'prepare':
                config_preparer = config_preparer.replace('prepare:', 'prepare_val:')

            elif self.mode == 'validation':
                config_preparer = config_preparer.replace('prepare:', 'prepare_val:')
                config_answers = config_answers.replace('/private/', '/private_val/')
                config_sample_submission = config_sample_submission.replace('/public/', '/public_val/')
                public_folder = 'public_val'
                private_folder = 'private_val'

            # Some benchmarks may not provide *_val splits; if missing, fall back to test artifacts.
            if self.mode == "validation":
                data_dir = self.get_data_dir()
                answers_candidate = data_dir / config_answers
                sample_candidate = data_dir / config_sample_submission
                public_candidate = data_dir / competition_id / "prepared" / public_folder
                private_candidate = data_dir / competition_id / "prepared" / private_folder
                if not (
                    answers_candidate.exists()
                    and sample_candidate.exists()
                    and public_candidate.exists()
                    and private_candidate.exists()
                ):
                    config_preparer = base_preparer
                    config_answers = base_answers
                    config_sample_submission = base_sample_submission
                    public_folder = "public"
                    private_folder = "private"

        # DABench competitions are not importable as Python packages (hyphenated ids).
        # Convert legacy import strings to file-based imports under the resolved root_dir.
        if is_dabench:
            module_str, fn_name = config_preparer.split(":")
            leaf = module_str.split(".")[-1]  # prepare or grade
            file_module = root_dir / competition_id / f"{leaf}.py"
            if not file_module.exists() and leaf.endswith("_val"):
                fallback = root_dir / competition_id / f"{leaf[:-4]}.py"
                if fallback.exists():
                    file_module = fallback
            config_preparer = f"file:{file_module}:{fn_name}"
            if "grader" in config and "grade_fn" in config["grader"]:
                g_module_str, g_fn_name = config["grader"]["grade_fn"].split(":")
                g_leaf = g_module_str.split(".")[-1]
                g_file_module = root_dir / competition_id / f"{g_leaf}.py"
                if not g_file_module.exists() and g_leaf.endswith("_val"):
                    g_fallback = root_dir / competition_id / f"{g_leaf[:-4]}.py"
                    if g_fallback.exists():
                        g_file_module = g_fallback
                config["grader"]["grade_fn"] = f"file:{g_file_module}:{g_fn_name}"
        else:
            config_preparer = self._coerce_file_import(
                config_preparer, root_dir, competition_id
            )
            if "grader" in config and "grade_fn" in config["grader"]:
                config["grader"]["grade_fn"] = self._coerce_file_import(
                    config["grader"]["grade_fn"], root_dir, competition_id
                )

        if is_dabench:
            def preparer_fn(raw: Path, public: Path, private: Path) -> Path:
                logger.info(
                    "DABench prepare disabled; using existing prepared/public + prepared/private."
                )
                return public
        else:
            preparer_fn = import_fn(config_preparer)

        answers = self.get_data_dir() / config_answers
        gold_submission = answers
        if "gold_submission" in config["dataset"]:
            gold_submission = self.get_data_dir() / config["dataset"]["gold_submission"]
        sample_submission = self.get_data_dir() / config_sample_submission

        raw_dir = self.get_data_dir() / competition_id / "raw"
        private_dir = self.get_data_dir() / competition_id / "prepared" / private_folder
        public_dir = self.get_data_dir() / competition_id / "prepared" / public_folder

        return Competition.from_dict(
            {
                **config,
                "description": description,
                "answers": answers,
                "sample_submission": sample_submission,
                "gold_submission": gold_submission,
                "prepare_fn": preparer_fn,
                "raw_dir": raw_dir,
                "private_dir": private_dir,
                "public_dir": public_dir,
                "checksums": checksums_path,
                "leaderboard": leaderboard_path,
            }
        )

    def get_competitions_dir(self) -> Path:
        """Retrieves the competition directory within the registry."""

        return get_module_dir() / "competitions"

    def get_splits_dir(self) -> Path:
        """Retrieves the splits directory within the repository."""

        return get_repo_dir() / "experiments" / "splits"

    def get_lite_competition_ids(self) -> list[str]:
        """List all competition IDs for the lite version (low complexity competitions)."""

        lite_competitions_file = self.get_splits_dir() / "low.txt"
        with open(lite_competitions_file, "r") as f:
            competition_ids = f.read().splitlines()
        return competition_ids

    def get_data_dir(self) -> Path:
        """Retrieves the data directory within the registry."""

        return self._data_dir

    def set_data_dir(self, new_data_dir: Path) -> "Registry":
        """Sets the data directory within the registry."""

        return Registry(new_data_dir)

    def list_competition_ids(self) -> list[str]:
        """List all competition IDs available in the registry, sorted alphabetically."""

        repo_dir = get_repo_dir()
        dabench_root = repo_dir / "dabench" / "competitions"

        search_roots = [repo_dir / "benchmarks" / "mlebench" / "competitions"]
        if dabench_root.exists():
            search_roots.append(dabench_root)

        competition_ids: set[str] = set()
        for root in search_roots:
            for cfg in root.rglob("config.yaml"):
                competition_ids.add(cfg.parent.stem)

        return sorted(competition_ids)


registry = Registry()



import json
import random
import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import get_logger

logger = get_logger(__name__)


def _organize_data_by_category(metadata: dict) -> dict:
    """Organizes image and annotation data by category ID."""
    annotations_images_by_category = {}
    for annotation, image in list(zip(metadata["annotations"], metadata["images"])):
        assert annotation["image_id"] == image["image_id"]
        category_id = annotation["category_id"]
        if category_id not in annotations_images_by_category:
            annotations_images_by_category[category_id] = []
        annotations_images_by_category[category_id].append(
            {
                "annotation": annotation,
                "image": image,
            }
        )
    return annotations_images_by_category


def _split_data(data_by_category: dict, test_size: float, random_state: int) -> tuple[dict, dict]:
    """Splits data for each category into train and test sets."""
    train_annotations_images_by_category = {}
    test_annotations_images_by_category = {}
    for category_id, annotations_images in data_by_category.items():
        # Each category needs to be in both train and test
        train_annotations_images, test_annotations_images = train_test_split(
            annotations_images, test_size=test_size, random_state=random_state
        )
        assert len(train_annotations_images) > 0 and len(test_annotations_images) > 0
        train_annotations_images_by_category[category_id] = train_annotations_images
        test_annotations_images_by_category[category_id] = test_annotations_images
    return train_annotations_images_by_category, test_annotations_images_by_category


def _process_train_set(
    train_data: dict, base_metadata: dict, raw_path: Path, output_public_path: Path
):
    """Processes and writes the training set data, images, and metadata."""
    logger.info(f"Processing train set for output: {output_public_path}")
    new_train_metadata = base_metadata.copy()
    new_train_metadata.update({"annotations": [], "images": []})
    train_sample_count = sum(len(v) for v in train_data.values())

    output_train_images_path = output_public_path / "train_images"

    with tqdm(
        desc=f"Creating train dataset in {output_public_path.name}",
        total=train_sample_count,
    ) as pbar:
        for category_id, annotations_images in train_data.items():
            category_subdir = f"{category_id // 100:03d}/{category_id % 100:02d}"
            (output_train_images_path / category_subdir).mkdir(exist_ok=True, parents=True)
            for idx, annotation_image in enumerate(annotations_images):
                new_image_id = f"{category_id:05d}__{(idx + 1):03d}"
                new_file_name = f"{category_subdir}/{new_image_id}.jpg"

                new_annotation = annotation_image["annotation"].copy()
                new_annotation["image_id"] = new_image_id
                new_train_metadata["annotations"].append(new_annotation)

                new_image = annotation_image["image"].copy()
                new_image["image_id"] = new_image_id
                new_image["file_name"] = new_file_name
                new_train_metadata["images"].append(new_image)

                src_path = raw_path / "train_images" / annotation_image["image"]["file_name"]
                dst_path = output_train_images_path / new_file_name
                shutil.copyfile(src=src_path, dst=dst_path)

                pbar.update(1)

    with open(output_public_path / "train_metadata.json", "w") as f:
        json.dump(new_train_metadata, f, indent=4, sort_keys=True)

    assert len(list(output_train_images_path.glob("**/*.jpg"))) == len(
        new_train_metadata["images"]
    )
    assert len(new_train_metadata["annotations"]) == len(new_train_metadata["images"])


def _process_test_set(
    test_data: dict, raw_path: Path, output_public_path: Path, output_private_path: Path
):
    """Processes and writes the test set data, images, metadata, and private answers."""
    logger.info(
        f"Processing test set for outputs: {output_public_path} and {output_private_path}"
    )
    new_test_metadata = {"annotations": [], "images": []}
    test_annotations_images = [item for sublist in test_data.values() for item in sublist]
    random.Random(0).shuffle(test_annotations_images)

    output_test_images_path = output_public_path / "test_images"

    for idx, annotation_image in tqdm(
        enumerate(test_annotations_images),
        desc=f"Creating test dataset in {output_public_path.name}",
        total=len(test_annotations_images),
    ):
        new_image_id = str(idx)
        new_file_name = f"{idx // 1000:03d}/test-{idx:06d}.jpg"

        new_annotation = annotation_image["annotation"].copy()
        new_annotation["image_id"] = new_image_id
        new_test_metadata["annotations"].append(new_annotation)

        new_image = annotation_image["image"].copy()
        new_image["image_id"] = new_image_id
        new_image["file_name"] = new_file_name
        new_test_metadata["images"].append(new_image)

        src_path = raw_path / "train_images" / annotation_image["image"]["file_name"]
        dst_path = output_test_images_path / new_file_name
        dst_path.parent.mkdir(exist_ok=True, parents=True)
        shutil.copyfile(src=src_path, dst=dst_path)

    with open(output_public_path / "test_metadata.json", "w") as f:
        json.dump(new_test_metadata["images"], f, indent=4, sort_keys=True)

    answers_rows = [
        {"Id": image["image_id"], "Predicted": annotation["category_id"]}
        for image, annotation in zip(new_test_metadata["images"], new_test_metadata["annotations"])
    ]
    answers_df = pd.DataFrame(answers_rows)
    answers_df.to_csv(output_private_path / "answers.csv", index=False)

    sample_rows = [{"Id": image["image_id"], "Predicted": 42} for image in new_test_metadata["images"]]
    sample_df = pd.DataFrame(sample_rows)
    sample_df.to_csv(output_public_path / "sample_submission.csv", index=False)

    assert len(list(output_test_images_path.glob("**/*.jpg"))) == len(new_test_metadata["images"])
    assert len(new_test_metadata["annotations"]) == len(new_test_metadata["images"])
    assert len(answers_df) == len(new_test_metadata["images"])
    assert len(sample_df) == len(answers_df)


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the raw data into public and private datasets with appropriate test/train splits.

    `train_metadata.json` is the "table of contents" for our data, with the following structure:
    (More details at https://www.kaggle.com/competitions/herbarium-2022-fgvc9/data)
    ```
    {
        "annotations" : [annotation],
        "categories" : [category],
        "genera" : [genus]
        "images" : [image],
        "distances" : [distance],
        "licenses" : [license],
        "institutions" : [institution]
    }
    ```
    - `images` and `annotations` are both N-length lists corresponding to the N samples.
        We'll need to split each of these lists into train and test.
    - The other fields are dataset-wide metadata that we don't need to touch.

    Other notes:
    - train/test splits need to occur per category (each category should be in both train and test).
    - The `test_images` and `train_images` folders have nested subdirs to make it easier to browse
        - `train_images` is structured as `{category_id[:3]}/{category_id[3:]}/{image_id}.jpg`
        - `test_images` is structured as `{image_idx[:3]}/test-{image_idx}.jpg` (to not reveal the category)
    - When we create the new splits, we re-assign image indices so that we don't give away labels based on the index
        - train images are indexed within their own category
        - test images follow a flat index after shuffling the categories
    """
    # Load raw data and organize it by category
    with open(raw / "train_metadata.json") as f:
        old_train_metadata = json.load(f)
    annotations_images_by_category = _organize_data_by_category(old_train_metadata)

    # --- 1. Create the original public/private datasets ---
    # This first split creates the main train and test sets.
    # The outputs in `public` and `private` will be identical to the original script.
    logger.info("--- Creating original train/test split for 'public' and 'private' directories ---")
    original_train_split, original_test_split = _split_data(
        annotations_images_by_category, test_size=0.2, random_state=0
    )

    _process_train_set(original_train_split, old_train_metadata, raw, public)
    _process_test_set(original_test_split, raw, public, private)
    logger.info("Finished creating original 'public' and 'private' datasets.")

    # --- 2. Create the validation datasets from the original training set ---
    # This second split takes the `original_train_split` and splits it *again*
    # to create a new, smaller training set and a validation set.
    # The outputs are saved to new `public_val` and `private_val` directories.
    logger.info("--- Creating validation split for 'public_val' and 'private_val' directories ---")
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"
    public_val.mkdir(exist_ok=True, parents=True)
    private_val.mkdir(exist_ok=True, parents=True)

    # The input for this split is the *train set* from the first split.
    # We use the exact same splitting logic to ensure consistency.
    train_val_split, test_val_split = _split_data(
        original_train_split, test_size=0.2, random_state=0
    )

    _process_train_set(train_val_split, old_train_metadata, raw, public_val)
    _process_test_set(test_val_split, raw, public_val, private_val)
    logger.info("Finished creating validation 'public_val' and 'private_val' datasets.")


import glob
import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.utils import extract, read_csv


def _process_split(
    source_df: pd.DataFrame,
    test_size: float,
    random_state: int,
    source_geometry_paths: list,
    public_dir: Path,
    private_dir: Path,
) -> pd.DataFrame:
    """
    Helper function to perform a data split, re-index, and write all necessary files.

    Args:
        source_df: The DataFrame to split.
        test_size: The proportion of the dataset to allocate to the test split.
        random_state: The seed used by the random number generator.
        source_geometry_paths: A list of paths to all original geometry files.
        public_dir: The destination directory for public-facing files (train set, test features).
        private_dir: The destination directory for private-facing files (test labels).

    Returns:
        The created training DataFrame, which can be used for a subsequent split.
    """
    # Ensure destination directories exist
    public_dir.mkdir(parents=True, exist_ok=True)
    private_dir.mkdir(parents=True, exist_ok=True)

    # Create train, test from the source dataframe
    new_train, new_test = train_test_split(
        source_df, test_size=test_size, random_state=random_state
    )

    # Make ids go 1, 2, ... for both train and test. Keep old ids so we can map ids of other files
    old_train_id_to_new = {
        old_id: new_id for new_id, old_id in enumerate(new_train["id"], start=1)
    }  # id starts from 1
    new_train["id"] = new_train["id"].map(old_train_id_to_new)

    old_test_id_to_new = {
        old_id: new_id for new_id, old_id in enumerate(new_test["id"], start=1)
    }  # id starts from 1
    new_test["id"] = new_test["id"].map(old_test_id_to_new)

    new_test_without_labels = new_test.drop(
        columns=["formation_energy_ev_natom", "bandgap_energy_ev"]
    )

    # Copy over files
    new_train.to_csv(public_dir / "train.csv", index=False)
    new_test.to_csv(private_dir / "test.csv", index=False)
    new_test_without_labels.to_csv(public_dir / "test.csv", index=False)

    # --- Process and copy geometry files for the new train set ---
    train_geometry_dir = public_dir / "train"
    for src in source_geometry_paths:
        original_id = int(Path(src).parts[-2])
        if original_id not in old_train_id_to_new:  # Filter for train ids
            continue

        new_id = old_train_id_to_new[original_id]
        dest_dir = train_geometry_dir / str(new_id)
        dest_dir.mkdir(parents=True, exist_ok=True)
        shutil.copy(src=src, dst=dest_dir / "geometry.xyz")
    assert len(list(train_geometry_dir.glob("**/*.xyz"))) == len(
        new_train
    ), f"Expected {len(new_train)} train geometry files in {public_dir}, found {len(list(train_geometry_dir.glob('**/*.xyz')))}"

    # --- Process and copy geometry files for the new test set ---
    test_geometry_dir = public_dir / "test"
    for src in source_geometry_paths:
        original_id = int(Path(src).parts[-2])
        if original_id not in old_test_id_to_new:  # Filter for test ids
            continue

        new_id = old_test_id_to_new[original_id]
        dest_dir = test_geometry_dir / str(new_id)
        dest_dir.mkdir(parents=True, exist_ok=True)
        shutil.copy(src=src, dst=dest_dir / "geometry.xyz")
    assert len(list(test_geometry_dir.glob("**/*.xyz"))) == len(
        new_test
    ), f"Expected {len(new_test)} test geometry files in {public_dir}, found {len(list(test_geometry_dir.glob('**/*.xyz')))}"

    # Create mock submission
    sample_submission = pd.DataFrame(
        {"id": new_test["id"], "formation_energy_ev_natom": 0.1779, "bandgap_energy_ev": 1.8892}
    )
    sample_submission.to_csv(public_dir / "sample_submission.csv", index=False)
    assert len(sample_submission) == len(
        new_test
    ), f"Sample submission in {public_dir} should have the same number of rows as its test set"

    return new_train


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split (public_val/private_val) from the training data.
    """
    # Define paths for the new validation set
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Extract only what we need
    extract(raw / "train.zip", raw / "train")
    extract(raw / "train.csv.zip", raw / "train.csv")
    extract(raw / "test.zip", raw / "test")
    extract(raw / "test.csv.zip", raw / "test.csv")

    # Load initial data and geometry paths
    initial_train_df = read_csv(raw / "train.csv/train.csv")
    all_geometry_paths = sorted(glob.glob(str(raw / "train/train/**/*.xyz")))

    # --- First Split: Create original public/private datasets ---
    # This split is identical to the original script to ensure unchanged outputs.
    train_df_after_first_split = _process_split(
        source_df=initial_train_df,
        test_size=0.1,
        random_state=0,
        source_geometry_paths=all_geometry_paths,
        public_dir=public,
        private_dir=private,
    )

    # --- Second Split: Create new validation datasets from the first split's train set ---
    # The test_size is set to 1/9 of the incoming data (which is 90% of the original).
    # This makes the new test_val set size equal to the original test set size (0.1 * T = 1/9 * 0.9 * T).
    _process_split(
        source_df=train_df_after_first_split,
        test_size=1 / 9,
        random_state=0,
        source_geometry_paths=all_geometry_paths,
        public_dir=public_val,
        private_dir=private_val,
    )


from pathlib import Path

from mlebench.utils import extract, read_csv

from .classes import CLASSES


def prepare(raw: Path, public: Path, private: Path):
    # extract only what is needed
    extract(raw / "train.csv.zip", raw)
    extract(raw / "test.csv.zip", raw)
    extract(raw / "test_labels.csv.zip", raw)
    extract(raw / "sample_submission.csv.zip", raw)

    # the test set is provided, so we dont have to split the train set nor form the sample submission
    train_with_labels = read_csv(raw / "train.csv")
    test_without_labels = read_csv(raw / "test.csv")
    answers = read_csv(raw / "test_labels.csv")
    sample_submission = read_csv(raw / "sample_submission.csv")
    sample_submission[CLASSES] = 0.5

    # save to public
    train_with_labels.to_csv(public / "train.csv", index=False)
    test_without_labels.to_csv(public / "test.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    # save to private
    answers.to_csv(private / "test.csv", index=False)

    assert len(answers) == len(
        sample_submission
    ), "Private test set and sample submission should be of the same length"

    assert sorted(answers["id"]) == sorted(
        test_without_labels["id"]
    ), "Private and Public test IDs should match"
    assert sorted(sample_submission["id"]) == sorted(
        test_without_labels["id"]
    ), "Public test and sample submission IDs should match"
    assert (
        len(set(train_with_labels["id"]) & set(test_without_labels["id"])) == 0
    ), "Train and test IDs should not overlap"



import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_log_error

from mlebench.grade_helpers import InvalidSubmissionError


TARGET_COLUMNS = ['num_views', 'num_votes', 'num_comments']


def prepare_for_metric(submission: pd.DataFrame, answers: pd.DataFrame):
    """Prepare submission and answers for RMSLE calculation."""
    if len(submission) != len(answers):
        raise InvalidSubmissionError(
            f"Submission length ({len(submission)}) != answers length ({len(answers)})"
        )

    if "id" not in submission.columns:
        raise InvalidSubmissionError("Submission must have an 'id' column")

    for col in TARGET_COLUMNS:
        if col not in submission.columns:
            raise InvalidSubmissionError(f"Submission must have a '{col}' column")

    # Sort by id
    submission = submission.sort_values("id").reset_index(drop=True)
    answers = answers.sort_values("id").reset_index(drop=True)

    # Check id alignment
    if not (submission["id"] == answers["id"]).all():
        raise InvalidSubmissionError("Submission and answers id mismatch")

    # Validate predictions
    for col in TARGET_COLUMNS:
        try:
            submission[col] = submission[col].astype(float)
        except (ValueError, TypeError):
            raise InvalidSubmissionError(f"'{col}' predictions must be numeric")

        if submission[col].isnull().any():
            raise InvalidSubmissionError(f"'{col}' predictions cannot contain NaN values")

        if (submission[col] < 0).any():
            raise InvalidSubmissionError(f"'{col}' predictions cannot be negative")

    return submission, answers


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    """Calculate RMSLE score across all target columns."""
    submission, answers = prepare_for_metric(submission, answers)

    # Concatenate all predictions and true values
    all_pred = []
    all_true = []
    for col in TARGET_COLUMNS:
        all_pred.extend(submission[col].values)
        all_true.extend(answers[col].values)

    all_pred = np.array(all_pred)
    all_true = np.array(all_true)

    # RMSLE
    rmsle = np.sqrt(mean_squared_log_error(all_true, all_pred))

    return rmsle



import shutil
from dataclasses import dataclass
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import extract

CLASSES = [
    "yes",
    "no",
    "up",
    "down",
    "left",
    "right",
    "on",
    "off",
    "stop",
    "go",
    "unknown",
    "silence",
]


@dataclass(frozen=True)
class AudioFile:
    label: str
    path: Path


def prepare(raw: Path, public: Path, private: Path):
    # extract only what we need
    extract(raw / "train.7z", raw)

    # Create train, test from train split
    audio_dir = raw / "train" / "audio"
    audio_files = sorted(
        [AudioFile(fpath.parent.name, fpath) for fpath in audio_dir.rglob("*.wav")],
        key=lambda x: f"{x.label}_{x.path.name}",
    )
    train_files, test_files = train_test_split(audio_files, test_size=0.1, random_state=0)

    # Make necessary directories
    labels = list(
        dict.fromkeys([file.label for file in train_files])
    )  # Gets unique elements deterministically

    for label in labels:
        (public / "train" / "audio" / label).mkdir(parents=True, exist_ok=True)

    (public / "test" / "audio").mkdir(parents=True, exist_ok=True)

    # Copy over train and test files
    for file in tqdm(train_files, desc="Copying train files"):
        shutil.copyfile(
            src=file.path,
            dst=public / "train" / "audio" / file.label / file.path.name,
        )

    test_records = []

    for idx, file in enumerate(tqdm(test_files, desc="Copying test files")):
        # Rename files, since training audio files across labels aren't necessarily unique.
        new_id = str(idx).zfill(8)
        new_name = f"clip_{new_id}.wav"
        test_records.append({"fname": new_name, "label": file.label})

        shutil.copyfile(
            src=file.path,
            dst=public / "test" / "audio" / new_name,
        )

    test = pd.DataFrame.from_records(test_records)
    test.to_csv(private / "test.csv", index=False)

    test_without_labels = test.drop(columns=["label"])
    sample_submission = test_without_labels.copy()
    sample_submission["label"] = "silence"
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    # Sanity checks
    test_audio_files = list((public / "test" / "audio").glob("*.wav"))
    num_test_files = len(test_audio_files)
    num_submission_entries = len(sample_submission)
    assert num_test_files == num_submission_entries, (
        f"The number of test audio files ({num_test_files}) does not match the number of entries "
        f"in sample_submission.csv ({num_submission_entries}). Please check the file copying process."
    )



import pandas as pd
from sklearn.metrics import f1_score

from mlebench.grade_helpers import InvalidSubmissionError


def prepare_for_metric(
    submission: pd.DataFrame,
    answers: pd.DataFrame,
    id_col: str = "Id",
    target_col: str = "Category",
):
    """
    `submission` and `answers` are pd.DataFrame with `id_col` and `target_col` columns
    """

    # Answers checks
    assert id_col in answers.columns, f"Answers must have '{id_col}' column"
    assert target_col in answers.columns, f"Answers must have '{target_col}' column"

    # Submission checks
    if len(submission) != len(answers):
        raise InvalidSubmissionError(
            f"Submission length {len(submission)} != answers length {len(answers)}"
        )
    if id_col not in submission.columns:
        raise InvalidSubmissionError(f"Submission must have '{id_col}' column")
    if target_col not in submission.columns:
        raise InvalidSubmissionError(f"Submission must have '{target_col}' column")

    # Sort submission and answers by "Id"
    submission = submission.sort_values(id_col)
    answers = answers.sort_values(id_col)
    if (submission[id_col].values != answers[id_col].values).any():
        raise InvalidSubmissionError(f"Submission and answers have mismatched '{id_col}' columns")

    y_true = [int(y) for y in answers[target_col]]
    y_pred = [int(y) for y in submission[target_col]]
    return y_true, y_pred


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    y_true, y_pred = prepare_for_metric(submission, answers)
    return f1_score(y_true=y_true, y_pred=y_pred, average="macro")



from pathlib import Path
import pandas as pd

from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def _split_and_save(
    source_df: pd.DataFrame,
    test_size: float,
    random_state: int,
    public_dir: Path,
    private_dir: Path,
    label_col: str = "selected_text",
    id_col: str = "textID",
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Helper function to perform a train/test split and save the results.

    Args:
        source_df: The DataFrame to split.
        test_size: The proportion of the dataset to allocate to the test split.
        random_state: The seed used by the random number generator.
        public_dir: The directory to save public artifacts (train set, unlabeled test set).
        private_dir: The directory to save private artifacts (labeled test set).
        label_col: The name of the column containing the label.
        id_col: The name of the column containing the unique ID.

    Returns:
        A tuple containing the created train and test DataFrames.
    """
    # Ensure output directories exist
    public_dir.mkdir(parents=True, exist_ok=True)
    private_dir.mkdir(parents=True, exist_ok=True)

    # Perform the split
    train_df, test_df = train_test_split(
        source_df, test_size=test_size, random_state=random_state
    )
    test_df_without_labels = test_df.drop(columns=[label_col])

    # Save the datasets
    train_df.to_csv(public_dir / "train.csv", index=False)
    test_df.to_csv(private_dir / "test.csv", index=False)
    test_df_without_labels.to_csv(public_dir / "test.csv", index=False)

    assert len(test_df_without_labels) == len(
        test_df
    ), f"Expected test_df_without_labels ({len(test_df_without_labels)}) == test_df ({len(test_df)})"
    assert len(train_df) + len(test_df) == len(
        source_df
    ), f"Expected train_df ({len(train_df)}) + test_df ({len(test_df)}) == source_df ({len(source_df)})"

    # Create and save a sample submission file
    submission_df = test_df.copy()[[id_col, label_col]]
    submission_df[label_col] = ""

    submission_df.to_csv(public_dir / "sample_submission.csv", index=False)
    assert len(submission_df) == len(
        test_df
    ), f"Expected submission_df ({len(submission_df)}) == test_df ({len(test_df)})"

    return train_df, test_df


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a second, parallel split for validation purposes.
    """
    old_train = read_csv(raw / "train.csv")

    # ---- Original Data Split (for final test set) ----
    # This split creates the main train and test sets for the competition.
    # The outputs in `public` and `private` are left completely unchanged.
    train_for_val_split, _ = _split_and_save(
        source_df=old_train,
        test_size=0.1,
        random_state=0,
        public_dir=public,
        private_dir=private,
    )

    # ---- New Validation Data Split ----
    # This second split takes the training set created above and splits it *again*
    # to create a new, smaller training set and a validation set. The results are
    # saved to parallel `public_val` and `private_val` directories.

    # Define paths for the new validation split
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # The original test set size was 0.1 of the total data. The new train set
    # is 0.9 of the total. To make the new validation set (test_val) have
    # roughly the same number of samples as the original test set, we use a
    # test_size of (0.1 / 0.9) = 1/9.
    validation_test_size = 1 / 9

    _split_and_save(
        source_df=train_for_val_split,
        test_size=validation_test_size,
        random_state=0,  # Use same random_state for consistency
        public_dir=public_val,
        private_dir=private_val,
    )


import json
import shutil
from pathlib import Path

import pandas as pd
from tqdm.auto import tqdm

from mlebench.utils import read_csv


# Create train, test from train split
def create_train_test_split(train_ancestors_df: pd.DataFrame, test_size: int):
    """
    On Kaggle, a user may "fork" (that is, copy) the notebook of another user to create their own version.
    `train_ancestors_df` contains information about the ancestry of each notebook in the training set.

    To create the test split, we select rows from the train set that don't share ancestors with any others
    so that there aren't any relatives in the train set. The train split is the remaining rows.
    """
    group_by_ancestor = {}
    for _, row in tqdm(
        train_ancestors_df.iterrows(), desc="Grouping by Ancestor", total=len(train_ancestors_df)
    ):
        ancestor_id = row["ancestor_id"]
        if ancestor_id not in group_by_ancestor:
            group_by_ancestor[ancestor_id] = []
        group_by_ancestor[ancestor_id].append(row)

    train_ids = []
    test_ids = []
    num_test_ids = 0
    # Ensure a deterministic split by sorting the groups
    sorted_ancestors = sorted(group_by_ancestor.items(), key=lambda x: x[0])
    for ancestor_id, rows in tqdm(sorted_ancestors, desc="Splitting on Ancestors"):
        if num_test_ids + len(rows) <= test_size:
            test_ids.extend(([row["id"] for row in rows]))
            num_test_ids += len(rows)
        else:
            train_ids.extend([row["id"] for row in rows])

    # The exact number of test samples can vary slightly based on group sizes.
    # We adjust the assertion to check it's close to the target, not necessarily equal.
    assert len(train_ids) + len(test_ids) == len(train_ancestors_df)

    return train_ids, test_ids


def _generate_split_files(
    train_ids: list,
    test_ids: list,
    public_dir: Path,
    private_dir: Path,
    raw_dir: Path,
    full_train_orders_df: pd.DataFrame,
    full_train_ancestors_df: pd.DataFrame,
):
    """
    A helper function to generate all required files for a given train/test split.
    This helps avoid code duplication and ensures consistent output structure.
    """
    # Create output directories
    public_dir.mkdir(parents=True, exist_ok=True)
    private_dir.mkdir(parents=True, exist_ok=True)
    (public_dir / "train").mkdir(exist_ok=True)
    (public_dir / "test").mkdir(exist_ok=True)

    # Copy json files to public
    for train_id in tqdm(train_ids, desc=f"Copying train json files to {public_dir.name}"):
        shutil.copy(raw_dir / "train" / f"{train_id}.json", public_dir / "train" / f"{train_id}.json")
    for test_id in tqdm(test_ids, desc=f"Copying test json files to {public_dir.name}"):
        shutil.copy(raw_dir / "train" / f"{test_id}.json", public_dir / "test" / f"{test_id}.json")

    # Generate answers for train and test
    # Answers for new train
    train_orders_new = full_train_orders_df[full_train_orders_df["id"].isin(train_ids)]
    train_orders_new.to_csv(public_dir / "train_orders.csv", index=False)
    # Answers for new test
    test_orders_new = full_train_orders_df[full_train_orders_df["id"].isin(test_ids)]
    test_orders_new.to_csv(private_dir / "test_orders.csv", index=False)

    # Make new train_ancestors.csv, excluding the new_test_ids
    train_ancestors_new = full_train_ancestors_df[
        full_train_ancestors_df["id"].isin(train_ids)
    ]
    train_ancestors_new.to_csv(public_dir / "train_ancestors.csv", index=False)

    # Create sample submission (use the given order without changing it)
    sample_submission_rows = []
    for sample_id in tqdm(
        test_orders_new["id"], desc=f"Creating sample submission for {public_dir.name}"
    ):
        # Get cell order from json file
        with open(public_dir / "test" / f"{sample_id}.json") as f:
            json_data = json.load(f)
            cell_order = list(json_data["cell_type"].keys())
        sample_submission_rows.append({"id": sample_id, "cell_order": " ".join(cell_order)})
    sample_submission = pd.DataFrame(sample_submission_rows)
    sample_submission.to_csv(public_dir / "sample_submission.csv", index=False)
    assert len(sample_submission) == len(test_ids), "Sample submission length does not match test length."


def prepare(raw: Path, public: Path, private: Path):
    # --- Initial Setup ---
    TEST_SIZE = 20000
    # Load all necessary data once
    full_train_ancestors_df = read_csv(raw / "train_ancestors.csv")
    full_train_orders_df = read_csv(raw / "train_orders.csv")

    # Shuffle the train_ancestors_df to ensure our split is random but reproducible
    shuffled_ancestors_df = full_train_ancestors_df.sample(
        frac=1, random_state=0
    ).reset_index(drop=True)

    # --- 1. Original Split (train/test) ---
    print("--- Generating original train/test split ---")
    train_ids, test_ids = create_train_test_split(shuffled_ancestors_df, test_size=TEST_SIZE)

    # Generate files for the original public and private directories
    _generate_split_files(
        train_ids,
        test_ids,
        public,
        private,
        raw,
        full_train_orders_df,
        shuffled_ancestors_df,
    )

    # --- 2. Validation Split (train_val/test_val) ---
    print("\n--- Generating new validation split ---")
    # Define paths for the new validation directories
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # The input for the second split is the training set from the first split.
    # We use the already shuffled dataframe to maintain deterministic behavior.
    train_subset_for_val_split = shuffled_ancestors_df[
        shuffled_ancestors_df["id"].isin(train_ids)
    ].reset_index(drop=True)

    # Create the second split from the first split's training data
    train_val_ids, test_val_ids = create_train_test_split(
        train_subset_for_val_split, test_size=TEST_SIZE
    )

    # Generate the validation split files in the new directories
    _generate_split_files(
        train_val_ids,
        test_val_ids,
        public_val,
        private_val,
        raw,
        full_train_orders_df,
        shuffled_ancestors_df,
    )

    print("\nData preparation complete.")
    print(f"Original split created in '{public.name}' and '{private.name}'.")
    print(f"Validation split created in '{public_val.name}' and '{private_val.name}'.")


import numpy as np
import pandas as pd
from sklearn.metrics import root_mean_squared_error

from mlebench.grade_helpers import InvalidSubmissionError


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    if len(submission) != len(answers):
        raise InvalidSubmissionError(
            f"Expected submission to be the same length as answers, but got {len(submission)} "
            f"instead of {len(answers)}."
        )

    to_predict = ["reactivity", "deg_Mg_pH10", "deg_Mg_50C"]
    expected_answer_columns = ["id_seqpos"] + to_predict + ["keep"]

    assert set(answers.columns).issuperset(expected_answer_columns), (
        f"Expected answers to have columns {expected_answer_columns}, but instead it has "
        f"columns {answers.columns}."
    )

    # The submission csv contains two columns which aren't used for scoring: `deg_pH10` and
    # `deg_50C`. These are nonetheless still required to be in the submission as per the
    # competition rules. See the "Sample Submission" section of the competition overview page for
    # more information:
    # https://www.kaggle.com/competitions/stanford-covid-vaccine/overview/evaluation
    expected_submission_columns = ["id_seqpos"] + to_predict + ["deg_pH10", "deg_50C"]

    if not set(submission.columns).issuperset(expected_submission_columns):
        raise InvalidSubmissionError(
            f"Expected the submission to have columns {expected_submission_columns}, but instead "
            f"it has columns {submission.columns}."
        )

    filtered_submission = submission[expected_submission_columns]

    # Sort rows by `id_seqpos` and columns alphabetically
    sorted_submission = filtered_submission.sort_values(by="id_seqpos").sort_index(axis=1)
    sorted_answers = answers.sort_values(by="id_seqpos").sort_index(axis=1)

    for i, (actual_id, expected_id) in enumerate(
        zip(sorted_submission["id_seqpos"], sorted_answers["id_seqpos"])
    ):
        if actual_id == expected_id:
            continue

        raise InvalidSubmissionError(
            f"Expected submission to have the same `id_seqpos` as answers, but got `{actual_id}` "
            f"instead of `{expected_id}` on row {i} of the submission."
        )

    mask = sorted_answers["keep"]
    new_submission = sorted_submission[mask]
    new_answers = sorted_answers[mask]

    errors = []

    for column in to_predict:
        y_pred = new_submission[column]
        y_true = new_answers[column]
        error = root_mean_squared_error(y_true=y_true, y_pred=y_pred)
        errors.append(error)

    return np.mean(errors)



CLASSES = [
    "ETT - Abnormal",
    "ETT - Borderline",
    "ETT - Normal",
    "NGT - Abnormal",
    "NGT - Borderline",
    "NGT - Incompletely Imaged",
    "NGT - Normal",
    "CVC - Abnormal",
    "CVC - Borderline",
]



import os
import shutil
import zipfile
from pathlib import Path

import numpy as np
from tqdm.auto import tqdm

from mlebench.utils import extract, get_logger

np_rng = np.random.RandomState(0)

logger = get_logger(__name__)


def count_lines_in_file(file_path):
    line_count = 0
    with open(file_path, "r") as file:
        for _line in file:
            line_count += 1
    return line_count


def compress_file_to_zip(src_file: Path, zip_file: Path):
    with zipfile.ZipFile(zip_file, "w", zipfile.ZIP_DEFLATED) as zipf:
        zipf.write(src_file, arcname=src_file.name)


def remove_random_word(sentence: str) -> str:
    """
    Remove a random 'word' (sequence of characters, delimited by whitespace) from a sentence.
    Does not remove first or last words.

    Punctuation counts as a word, and is already separated by whitespace.
    """
    words = sentence.split()
    index = np_rng.randint(1, len(words) - 1)
    return " ".join(words[:index] + words[index + 1 :])


def prepare(raw: Path, public: Path, private: Path):
    logger.info("Extracting raw / train_v2.txt.zip")
    extract(raw / "train_v2.txt.zip", raw)

    # computed this ahead of time
    total_lines = 30301028

    with (
        open(raw / "train_v2.txt", "r") as old_train,
        open(public / "train_v2.txt", "w") as public_train,
        open(public / "test_v2.txt", "w") as public_test,
        open(private / "test.csv", "w") as private_test,
    ):
        public_test.write('"id","sentence"\n')
        private_test.write('"id","sentence"\n')
        line_count = 0
        test_count = 0
        train_count = 0
        # there is one sentence per line
        for sentence in tqdm(old_train, desc="Processing data", total=total_lines):
            # we will put ~0.01 of the data in test, the rest in train, matching kaggle's original split
            # some sentences only have 2 words, so can't remove a word -- keep them in train
            if np_rng.uniform() <= 0.01 and len(sentence.strip().split()) > 2:
                # get rid of linebreak and escape quotes
                sentence = sentence.strip().replace('"', '""')
                removed_word_sentence = remove_random_word(sentence)
                private_test.write(f'{test_count},"{sentence}"\n')
                public_test.write(f'{test_count},"{removed_word_sentence}"\n')
                test_count += 1
            else:
                public_train.write(sentence)
                train_count += 1
            line_count += 1
            if line_count >= total_lines:
                break

    # we will be compressing the public files (to match what's on kaggle.com)
    # so copy our sample submission to private so we have access to it
    shutil.copy(public / "test_v2.txt", private / "sample_submission.csv")

    # compress the public files
    logger.info("Compressing train_v2.txt")
    compress_file_to_zip(public / "train_v2.txt", public / "train_v2.txt.zip")
    logger.info("Compressing test_v2.txt")
    compress_file_to_zip(public / "test_v2.txt", public / "test_v2.txt.zip")
    # remove the original files
    (public / "train_v2.txt").unlink()
    (public / "test_v2.txt").unlink()

    # Checks
    assert not (public / "train_v2.txt").exists(), "public / 'train_v2.txt' should not exist"
    assert (public / "train_v2.txt.zip").exists(), "public / 'train_v2.txt.zip' should exist"
    assert not (public / "test_v2.txt").exists(), "public / 'test_v2.txt' should not exist"
    assert (public / "test_v2.txt.zip").exists(), "public / 'test_v2.txt.zip' should exist"

    private_test_line_count = count_lines_in_file(private / "test.csv")
    assert (
        # minus 2 to exclude header
        private_test_line_count - 1
        == test_count
    ), "private / 'test.csv' has incorrect number of lines"
    assert (
        count_lines_in_file(private / "sample_submission.csv") == private_test_line_count
    ), "private / 'sample_submission.csv' has incorrect number of lines"
    assert (
        test_count + train_count == total_lines
    ), "Expected the number of test samples and train samples to sum to the total number of samples in the original train file"



import random
import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    """
    # List of all train image IDs
    cover_images_dir = raw / "Cover"
    cover_images = sorted(list(cover_images_dir.glob("*.jpg")))

    # Create train, test from train split
    test_size = 5000
    new_train, new_test = train_test_split(cover_images, test_size=test_size, random_state=42)

    # Prepare dirs
    steganography_algs = ["JMiPOD", "JUNIWARD", "UERD"]
    for subdir in ["Cover", "Test"] + steganography_algs:
        (public / subdir).mkdir(parents=True, exist_ok=True)

    # Copy over new train set, giving them new ids
    for idx, fp in tqdm(enumerate(new_train), total=len(new_train), desc="Copying train images"):
        image_id = idx + 1  # 1-indexed
        shutil.copyfile(src=fp, dst=public / "Cover" / f"{image_id:05d}.jpg")
        for alg in steganography_algs:
            shutil.copyfile(src=raw / alg / fp.name, dst=public / alg / f"{image_id:05d}.jpg")

    # Populate test set
    answers_rows = []
    random.seed(0)
    random.shuffle(new_test)
    for idx, fp in tqdm(enumerate(new_test), total=len(new_test), desc="Copying test images"):
        image_id = idx + 1  # 1-indexed
        test_id = f"{image_id:04d}.jpg"
        dest = public / "Test" / test_id

        # For the test set, we randomly select between the "Cover" (unedited image, negative class)
        # and one of the 3 steganography algorithms (positive class)
        # 1:1 ratio of positive:negative examples, and even distribution of steganography algorithms
        if random.choice([True, False]):
            # Negative class
            shutil.copyfile(
                src=fp,
                dst=dest,
            )
            answers_rows.append({"Id": test_id, "Label": 0})
        else:
            # Positive class
            alg = random.choice(steganography_algs)
            shutil.copyfile(src=raw / alg / fp.name, dst=dest)
            answers_rows.append({"Id": test_id, "Label": 1})

    # Write answers to file
    answers_df = pd.DataFrame(answers_rows)
    answers_df.to_csv(private / "test.csv", index=False)

    # Create sample submission
    sample_submission = answers_df.copy()
    sample_submission["Label"] = 0
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    # Checks
    assert "Id" in answers_df.columns, "Answers must have 'Id' column"
    assert "Label" in answers_df.columns, "Answers must have 'Label' column"
    assert "Id" in sample_submission.columns, "Sample submission must have 'Id' column"
    assert "Label" in sample_submission.columns, "Sample submission must have 'Label' column"
    assert (
        len(answers_df) == test_size
    ), f"Expected {test_size} test images, but got {len(answers_df)}"
    assert len(sample_submission) == len(
        answers_df
    ), f"Sample submission ({len(sample_submission)}) and answers ({len(answers_df)}) must have the same length"
    assert (
        len(list(public.glob("Test/*.jpg"))) == test_size
    ), f"Expected {test_size} test images in public/Test, but got {len(list(public.glob('Test/*.jpg')))}"
    assert len(list(public.glob("Cover/*.jpg"))) == len(
        new_train
    ), f"Expected {len(new_train)} train images in public/Cover, but got {len(list(public.glob('Cover/*.jpg')))}"
    for alg in steganography_algs:
        assert len(list(public.glob(f"{alg}/*.jpg"))) == len(
            new_train
        ), f"Expected {len(new_train)} train images in public/{alg}, but got {len(list(public.glob(f'{alg}/*.jpg')))}"



CLASS_NAMES = [
    "Accelerating_and_revving_and_vroom",
    "Accordion",
    "Acoustic_guitar",
    "Applause",
    "Bark",
    "Bass_drum",
    "Bass_guitar",
    "Bathtub_(filling_or_washing)",
    "Bicycle_bell",
    "Burping_and_eructation",
    "Bus",
    "Buzz",
    "Car_passing_by",
    "Cheering",
    "Chewing_and_mastication",
    "Child_speech_and_kid_speaking",
    "Chink_and_clink",
    "Chirp_and_tweet",
    "Church_bell",
    "Clapping",
    "Computer_keyboard",
    "Crackle",
    "Cricket",
    "Crowd",
    "Cupboard_open_or_close",
    "Cutlery_and_silverware",
    "Dishes_and_pots_and_pans",
    "Drawer_open_or_close",
    "Drip",
    "Electric_guitar",
    "Fart",
    "Female_singing",
    "Female_speech_and_woman_speaking",
    "Fill_(with_liquid)",
    "Finger_snapping",
    "Frying_(food)",
    "Gasp",
    "Glockenspiel",
    "Gong",
    "Gurgling",
    "Harmonica",
    "Hi-hat",
    "Hiss",
    "Keys_jangling",
    "Knock",
    "Male_singing",
    "Male_speech_and_man_speaking",
    "Marimba_and_xylophone",
    "Mechanical_fan",
    "Meow",
    "Microwave_oven",
    "Motorcycle",
    "Printer",
    "Purr",
    "Race_car_and_auto_racing",
    "Raindrop",
    "Run",
    "Scissors",
    "Screaming",
    "Shatter",
    "Sigh",
    "Sink_(filling_or_washing)",
    "Skateboard",
    "Slam",
    "Sneeze",
    "Squeak",
    "Stream",
    "Strum",
    "Tap",
    "Tick-tock",
    "Toilet_flush",
    "Traffic_noise_and_roadway_noise",
    "Trickle_and_dribble",
    "Walk_and_footsteps",
    "Water_tap_and_faucet",
    "Waves_and_surf",
    "Whispering",
    "Writing",
    "Yell",
    "Zipper_(clothing)",
]
assert len(CLASS_NAMES) == 80, f"Expected 80 class names, but got {len(CLASS_NAMES)}"



import shutil
from itertools import islice
from pathlib import Path

import bson
import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    """

    dev_mode = False

    def read_ids_and_category_ids(file_path: Path) -> pd.DataFrame:
        data = bson.decode_file_iter(open(file_path, "rb"))

        records = []

        for c, d in enumerate(tqdm(data, desc="Reading BSON data")):
            records.append({"_id": d["_id"], "category_id": d["category_id"]})

        return pd.DataFrame(records)

    def filter_bson_by_ids(
        bson_file_path: Path,
        ids: set,
        write_path: Path,
        exclude_cols: list = [],
        chunk_size=1000,
        max_rows=None,
    ):
        """
        Filters a BSON file by a set of IDs and writes the filtered data to a new BSON file.
        The original _id is replaced with a new _id starting from 0 and incrementing by 1.

        Args:
            bson_file_path (Path): Path to the input BSON file.
            ids (set): Set of IDs to filter by.
            write_path (Path): Path to the output BSON file.
            exclude_cols (list): List of columns to exclude from the output.
            max_rows (int, optional): Maximum number of rows to write to the output file.
        """
        data = bson.decode_file_iter(open(bson_file_path, "rb"))
        num_written_rows = 0

        with open(write_path, "wb") as f:
            for record in tqdm(data, desc="Filtering BSON data"):
                if record["_id"] in ids:
                    for col in exclude_cols:
                        if col in record:
                            del record[col]
                    num_written_rows += 1
                    f.write(bson.BSON.encode(record))

                if num_written_rows % chunk_size == 0:
                    f.flush()

                if max_rows is not None and num_written_rows >= max_rows:
                    break

    # Create train, test from train split. Original train.bson contains 7,069,896 rows. Original test.bson contains 1,768,182 rows.
    old_train = read_ids_and_category_ids(raw / "train.bson")

    # Ensure rows in train_example remain in new_train
    new_train, answers = train_test_split(old_train, test_size=0.1, random_state=0)
    answers = answers.sort_values(by="_id")

    # Create sample submission
    sample_submission = answers[["_id"]]
    sample_submission["category_id"] = 1000010653

    # Checks
    assert len(new_train) + len(answers) == len(
        old_train
    ), f"The length of new_train and answers combined should be equal to the original length of old_train. Got {len(new_train) + len(answers)} and {len(old_train)}"
    assert set(new_train["_id"]).isdisjoint(
        set(answers["_id"])
    ), "new_train and answers should not have any _ids in common"
    assert sample_submission.columns.tolist() == [
        "_id",
        "category_id",
    ], f"sample_submission should have columns _id and category_id. Got {sample_submission.columns.tolist()}"

    # Write new files
    answers.to_csv(private / "answers.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    filter_bson_by_ids(
        bson_file_path=(
            raw / "train_example.bson" if dev_mode else raw / "train.bson"
        ),  # train_example.bson is the first 100 rows of train.bson
        ids=set(new_train["_id"]),
        write_path=public / "train.bson",
    )
    filter_bson_by_ids(
        bson_file_path=raw / "train_example.bson" if dev_mode else raw / "train.bson",
        ids=set(answers["_id"]),
        write_path=public / "test.bson",
        exclude_cols=["category_id"],  # removes category_id for test.bson
    )

    # Write new train_example.bson which is the first 100 rows of the new train.bson
    filter_bson_by_ids(
        bson_file_path=(
            raw / "train_example.bson" if dev_mode else raw / "train.bson"
        ),  # train_example.bson is the first 100 rows of train.bson
        ids=set(new_train["_id"]),
        write_path=public / "train_example.bson",
        max_rows=100,
    )

    def is_valid_bson_file(file_path: Path, chunk_size: int = 10000):
        try:
            with open(file_path, "rb") as f:
                data_iter = bson.decode_file_iter(f)
                for chunk in tqdm(
                    iter(lambda: list(islice(data_iter, chunk_size)), []),
                    desc=f"Validating {file_path.name}",
                ):
                    pd.DataFrame(chunk)  # Attempt to create a DataFrame from the chunk
        except Exception as e:
            return False

        return True

    # Check train.bson
    assert is_valid_bson_file(public / "train.bson"), f"Couldn't parse `train.bson` as a bson file!"

    # Check test.bson
    assert is_valid_bson_file(public / "test.bson"), f"Couldn't parse `test.bson` as a bson file!"

    # Copy over other files
    shutil.copy(raw / "category_names.csv", public / "category_names.csv")

    actual_new_train = read_ids_and_category_ids(public / "train.bson")
    actual_new_train_example = read_ids_and_category_ids(public / "train_example.bson")

    assert actual_new_train.iloc[:100].equals(
        actual_new_train_example
    ), f"The first 100 rows of `train.bson` should be the same as `train_example.bson`"



from pandas import DataFrame
from sklearn.metrics import root_mean_squared_error

from mlebench.grade_helpers import InvalidSubmissionError


def grade(submission: DataFrame, answers: DataFrame) -> float:
    assert "id" in answers.columns, "Answers must have an 'id' column."
    assert "value" in answers.columns, "Answers must have a 'value' column."

    try:
        answers["value"].astype(float)
    except ValueError as e:
        assert False, "Answers must have float values in the 'value' column."

    if "id" not in submission.columns:
        raise InvalidSubmissionError("Submission must have an 'id' column.")

    if "value" not in submission.columns:
        raise InvalidSubmissionError("Submission must have a 'value' column.")

    try:
        submission["value"].astype(float)
    except ValueError as e:
        raise InvalidSubmissionError(f"Submission must have float values in the 'value' column.")

    if len(submission) != len(answers):
        raise InvalidSubmissionError(
            f"Expected the submission to have {len(answers)} rows, but got {len(submission)}."
        )

    submission_sorted = submission.sort_values(by="id").sort_index(axis=1)
    answers_sorted = answers.sort_values(by="id").sort_index(axis=1)

    if (submission_sorted["id"].values != answers_sorted["id"].values).any():
        raise InvalidSubmissionError(
            "Expected the submission to have the same 'id' values as the answers, but they differ."
        )

    y_true = submission_sorted["value"]
    y_pred = answers_sorted["value"]
    score = root_mean_squared_error(y_true=y_true, y_pred=y_pred)

    return score



from pydantic import BaseModel, Field, ConfigDict
from typing import Literal, Dict, Any

# Define a controlled vocabulary for task types.
# The framework will choose appropriate TaskHandler based on this type.
TaskType = Literal["kaggle", "qa", "code", "datasci", "open_ended"]

# Define execution modes.
# 'standard_ml': Strict evaluation against ground truth (e.g., RMSE, Accuracy).
# 'open_ended': Exploratory tasks evaluated by LLM judge or artifact existence.
TaskMode = Literal["standard_ml", "open_ended"]


class TaskDefinition(BaseModel):
    """
    Standardized, serializable representation of any task in the DSAT framework.

    This is the "logical contract" for the framework to interact with external benchmarks.
    It encapsulates all information about a task so that DSATRunner and TaskHandler
    can understand and process it.
    """
    task_id: str = Field(
        description="Unique identifier for the task instance, e.g., 'house-prices-advanced-regression-techniques', 'gsm8k_train_001'."
    )
    task_type: TaskType = Field(
        description="General category of the task, used to select the correct TaskHandler for processing."
    )
    mode: TaskMode = Field(
        default="standard_ml",
        description="The execution mode for the task. Defaults to 'standard_ml' for backward compatibility."
    )
    payload: Dict[str, Any] = Field(
        description="A dictionary containing task-specific data."
    )

    model_config = ConfigDict(
        frozen = True  # Task definitions should be immutable after creation.
    )


import pandas as pd
import numpy as np
import sys
from pathlib import Path


def prepare(raw: Path, public: Path, private: Path):
    """
    Prepare the ethanol-concentration dataset for the benchmark.

    This function converts the .ts time series files to numpy arrays,
    so that public directory only contains data, not data loading code.

    Args:
        raw: Path to raw data directory (contains EthanolConcentration_TRAIN.ts and EthanolConcentration_TEST.ts)
        public: Path to public directory (visible to participants)
        private: Path to private directory (hidden from participants, used for grading)
    """
    # Load test data to extract labels for grading
    try:
        # Add raw directory to path temporarily to import dataset module
        sys.path.insert(0, str(raw))

        # Use the local dataset module to load data
        from dataset import get_dataset

        # Change to raw directory to load data
        import os
        original_dir = os.getcwd()
        os.chdir(str(raw))

        try:
            # Load train and test datasets
            X_train, y_train = get_dataset('TRAIN')
            X_test, y_test = get_dataset('TEST')

            print(f"Loaded training data: X_train.shape = {X_train.shape}, y_train.shape = {y_train.shape}")
            print(f"Loaded test data: X_test.shape = {X_test.shape}, y_test.shape = {y_test.shape}")
            print(f"Number of unique labels: {len(np.unique(y_train))}")

            # Save training data as numpy arrays in public directory
            np.save(public / "train_data.npy", X_train)
            np.save(public / "train_labels.npy", y_train.flatten())

            # Save test data (without labels) in public directory
            np.save(public / "test_data.npy", X_test)

            # Create test labels dataframe for grading (private)
            test_labels_df = pd.DataFrame(
                {"id": range(len(y_test)), "label": y_test.flatten()}
            )
            test_labels_df.to_csv(private / "test_labels.csv", index=False)

            # Create sample submission file
            sample_submission = pd.DataFrame(
                {"id": range(len(y_test)), "label": 0}  # Default to class 0
            )
            sample_submission.to_csv(public / "sample_submission.csv", index=False)

            print(f"Data preparation completed:")
            print(f"  - Training: {len(X_train)} samples")
            print(f"  - Test: {len(X_test)} samples")
            print(f"  - Sequence length: {X_train.shape[1]}")
            print(f"  - Feature dimension: {X_train.shape[2]}")
            print(f"  - Number of classes: {len(np.unique(y_train))}")

        finally:
            os.chdir(original_dir)
            sys.path.remove(str(raw))

    except Exception as e:
        print(f"Error loading test labels: {e}")
        import traceback
        traceback.print_exc()
        # Fallback: create dummy files if loading fails
        print("Creating dummy submission files...")
        dummy_df = pd.DataFrame({"id": [0], "label": [0]})
        dummy_df.to_csv(private / "test_labels.csv", index=False)
        dummy_df.to_csv(public / "sample_submission.csv", index=False)

    # Validation checks
    assert (public / "train_data.npy").exists(), "Training data should exist"
    assert (public / "train_labels.npy").exists(), "Training labels should exist"
    assert (public / "test_data.npy").exists(), "Test data should exist"
    assert (private / "test_labels.csv").exists(), "Test labels should exist"
    assert (public / "sample_submission.csv").exists(), "Sample submission should exist"

    print(f"\nPrepared ethanol-concentration dataset:")
    print(f"  - Public files: {list(public.glob('*'))}")
    print(f"  - Private files: {list(private.glob('*'))}")



"""High-level grading functionality"""
import json
from datetime import datetime
from pathlib import Path

import pandas as pd
from tqdm import tqdm

from mlebench.data import get_leaderboard, is_dataset_prepared
from mlebench.grade_helpers import CompetitionReport
from mlebench.registry import Competition, Registry
from mlebench.registry import registry as DEFAULT_REGISTRY
from mlebench.utils import get_logger, get_timestamp, load_answers, purple, read_csv, read_jsonl

logger = get_logger(__name__)


def grade_jsonl(
    path_to_submissions: Path,
    output_dir: Path,
    registry: Registry = DEFAULT_REGISTRY,
):
    """
    Grades multiple submissions stored in a JSONL file.
    Saves the aggregated report as a JSON file.
    """

    submissions = read_jsonl(path_to_submissions, skip_commented_out_lines=True)
    competitions_reports = []

    for submission in tqdm(submissions, desc="Grading submissions", unit="submission"):
        submission_path = Path(str(submission["submission_path"]))
        competition_id = submission["competition_id"]
        competition = registry.get_competition(competition_id)
        single_report = grade_csv(submission_path, competition)
        competitions_reports.append(single_report)

    aggregated_report = aggregate_reports(competitions_reports)
    timestamp = get_timestamp()
    save_path = output_dir / f"{timestamp}_grading_report.json"
    logger.info(
        json.dumps(
            {k: v for k, v in aggregated_report.items() if k != "competition_reports"}, indent=4
        )
    )

    output_dir.mkdir(exist_ok=True)
    with open(save_path, "w") as f:
        json.dump(aggregated_report, f, indent=2)
    logger.info(purple(f"Saved summary report to {save_path}"))


def grade_csv(path_to_submission: Path, competition: Competition) -> CompetitionReport:
    """Grades a submission CSV for the given competition."""

    if not is_dataset_prepared(competition, grading_only=True):
        raise ValueError(
            f"Dataset for competition `{competition.id}` is not prepared! "
            f"Please run `mlebench prepare -c {competition.id}` to prepare the dataset."
        )

    score = None
    submission_exists = path_to_submission.is_file() and path_to_submission.suffix.lower() == ".csv"

    if submission_exists:
        submission_df = read_csv(path_to_submission)
        logger.info(purple(f"Load answers from {competition.answers}"))
        answers = load_answers(competition.answers)
        score = competition.grader(submission_df, answers)
    else:
        logger.warning(
            f"Invalid submission file: {path_to_submission}. Please check that the file exists and it is a CSV."
        )

    valid_submission = score is not None

    # Helper function to determine is_lower_better from metric name
    def infer_is_lower_better(metric_name: str) -> bool:
        """
        Infer whether lower is better based on metric name.
        Returns True if lower is better (error metrics), False if higher is better (accuracy metrics).
        """
        metric_name_lower = metric_name.lower()

        # Error/loss metrics (lower is better)
        error_metrics = [
            'rmse', 'rmsle', 'mae', 'mse', 'mape',
            'log_loss', 'crossentropy', 'kld', 'error',
            'loss', 'distance', 'deviation'
        ]

        # Accuracy/score metrics (higher is better)
        accuracy_metrics = [
            'accuracy', 'precision', 'recall', 'f1',
            'auc', 'roc', 'score', 'r2', 'correlation',
            'iou', 'dice', 'map', 'ndcg'
        ]

        for error_metric in error_metrics:
            if error_metric in metric_name_lower:
                return True

        for accuracy_metric in accuracy_metrics:
            if accuracy_metric in metric_name_lower:
                return False

        # Default: assume lower is better for safety
        return True

    # Try to get leaderboard, but make it optional
    try:
        competition_leaderboard = get_leaderboard(competition)
        rank_info = competition.grader.rank_score(score, competition_leaderboard)
        is_lower_better = competition.grader.is_lower_better(competition_leaderboard)
    except (AssertionError, FileNotFoundError):
        # Leaderboard not found, infer is_lower_better from grader name
        logger.warning(f"Leaderboard not found for competition {competition.id}, skipping rank calculation")
        competition_leaderboard = None
        rank_info = {
            "gold_threshold": None,
            "silver_threshold": None,
            "bronze_threshold": None,
            "median_threshold": None,
            "gold_medal": False,
            "silver_medal": False,
            "bronze_medal": False,
            "above_median": False,
        }
        # Infer is_lower_better from grader name
        is_lower_better = infer_is_lower_better(competition.grader.name)

    return CompetitionReport(
        competition_id=competition.id,
        score=score,
        gold_threshold=rank_info["gold_threshold"],
        silver_threshold=rank_info["silver_threshold"],
        bronze_threshold=rank_info["bronze_threshold"],
        median_threshold=rank_info["median_threshold"],
        any_medal=rank_info["gold_medal"] or rank_info["silver_medal"] or rank_info["bronze_medal"],
        gold_medal=rank_info["gold_medal"],
        silver_medal=rank_info["silver_medal"],
        bronze_medal=rank_info["bronze_medal"],
        above_median=rank_info["above_median"],
        submission_exists=submission_exists,
        valid_submission=valid_submission,
        is_lower_better=is_lower_better,
        created_at=datetime.now(),
        submission_path=str(path_to_submission),
    )


def validate_submission(submission: Path, competition: Competition) -> tuple[bool, str]:
    """
    Validates a submission for the given competition by actually running the competition grader.
    This is designed for end users, not developers (we assume that the competition grader is
    correctly implemented and use that for validating the submission, not the other way around).
    """
    if not submission.is_file():
        return False, f"Submission invalid! Submission file {submission} does not exist."

    if not submission.suffix.lower() == ".csv":
        return False, "Submission invalid! Submission file must be a CSV file."

    if not is_dataset_prepared(competition, grading_only=True):
        raise ValueError(
            f"Dataset for competition `{competition.id}` is not prepared! "
            f"Please run `mlebench prepare -c {competition.id}` to prepare the dataset."
        )

    try:
        competition.grader.grade_fn(read_csv(submission), read_csv(competition.answers))
    except Exception as e:
        return (
            False,
            f"Submission invalid! The attempt to grade the submission has resulted in the following error message:\n{e}",
        )

    return True, "Submission is valid."


def aggregate_reports(competition_reports: list[CompetitionReport]) -> dict:
    """
    Builds the summary report from a list of competition reports.
    If pass_at_n is True, then aggregate performence of competitions by selecting the best performance per
    competition, otherwise sum metrics
    """

    total_gold_medals = sum(report.gold_medal for report in competition_reports)
    total_silver_medals = sum(report.silver_medal for report in competition_reports)
    total_bronze_medals = sum(report.bronze_medal for report in competition_reports)
    total_above_median = sum(report.above_median for report in competition_reports)
    total_submissions = sum(report.submission_exists for report in competition_reports)
    total_valid_submissions = sum(report.valid_submission for report in competition_reports)

    summary_report = {
        "total_runs": int(len(competition_reports)),
        "total_runs_with_submissions": int(total_submissions),
        "total_valid_submissions": int(total_valid_submissions),
        "total_medals": int(total_gold_medals + total_silver_medals + total_bronze_medals),
        "total_gold_medals": int(total_gold_medals),
        "total_silver_medals": int(total_silver_medals),
        "total_bronze_medals": int(total_bronze_medals),
        "total_above_median": int(total_above_median),
        "competition_reports": [cr.to_dict() for cr in competition_reports],
    }

    return summary_report

def simple_accuracy_grader(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    """
    Computes simple accuracy between submission and answers.
    Assumes matching order or joins on an ID column.
    """
    try:
        # 1. Try to find ID column to merge on
        id_col = None
        common_cols = set(submission.columns).intersection(set(answers.columns))
        for col in common_cols:
            if col.lower() in ['id', 'key', 'index', 'sample_id', 'image_id', 'patient_id']:
                id_col = col
                break
                
        if id_col:
            merged = submission.merge(answers, on=id_col, suffixes=('_pred', '_true'))
            # Compare all other columns
            pred_cols = [c for c in submission.columns if c != id_col]
            true_cols = [c for c in answers.columns if c != id_col]
            
            # If there's only one other column, compare it
            if len(pred_cols) == 1 and len(true_cols) == 1:
                return float((merged[pred_cols[0]] == merged[true_cols[0]]).mean())
            
            # Iterate over shared columns that are NOT id
            shared_content_cols = [c for c in common_cols if c != id_col]
            if shared_content_cols:
                matches = []
                for c in shared_content_cols:
                    # In merged df, they are c_pred and c_true
                    matches.append((merged[f"{c}_pred"] == merged[f"{c}_true"]).mean())
                if matches:
                    return float(sum(matches) / len(matches))

        # 2. Fallback: Position-based comparison (assuming sorted)
        if len(submission) == len(answers):
            # Compare last column if dimensions match
            if submission.shape[1] == answers.shape[1]:
                return float((submission.iloc[:, -1] == answers.iloc[:, -1]).mean())
            
            # Or compare specific common columns
            common_cols = set(submission.columns).intersection(set(answers.columns))
            if common_cols:
                matches = []
                for c in common_cols:
                    matches.append((submission[c] == answers[c]).mean())
                if matches:
                    return float(sum(matches) / len(matches))
    except Exception as e:
        logger.error(f"simple_accuracy_grader failed: {e}")
        return 0.0

    return 0.0


import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def make_image_subpath(image_id: str) -> Path:
    """
    Creates a triple-nested directory structure from the first 3 characters of the image_id.
    """
    subpath = Path(image_id[0]) / image_id[1] / image_id[2] / f"{image_id}.png"
    return subpath


def _create_split_files(
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
    raw_images_path: Path,
    public_path: Path,
    private_path: Path,
):
    """
    Helper function to generate the directory structure and files for a given split.
    This function populates the public and private directories with train/test data,
    images, and a sample submission file.
    """
    # Create output directories
    public_path.mkdir(exist_ok=True)
    private_path.mkdir(exist_ok=True)

    # Save dataframes
    train_df.to_csv(public_path / "train_labels.csv", index=False)
    test_df.to_csv(private_path / "test.csv", index=False)

    # Copy train files
    desc_prefix = public_path.name
    for _, row in tqdm(train_df.iterrows(), total=len(train_df), desc=f"Copying {desc_prefix} train images"):
        image_id = row["image_id"]
        src = raw_images_path / make_image_subpath(image_id)
        dst = public_path / "train" / make_image_subpath(image_id)
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(src=src, dst=dst)

    # Copy test files
    for _, row in tqdm(test_df.iterrows(), total=len(test_df), desc=f"Copying {desc_prefix} test images"):
        image_id = row["image_id"]
        src = raw_images_path / make_image_subpath(image_id)
        dst = public_path / "test" / make_image_subpath(image_id)
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(src=src, dst=dst)

    # Create sample submission
    sample_submission = test_df.copy()
    sample_submission["InChI"] = "InChI=1S/H2O/h1H2"
    sample_submission.to_csv(public_path / "sample_submission.csv", index=False)

    # Checks
    assert len(list((public_path / "train").glob("**/*.png"))) == len(
        train_df
    ), f"Expected {len(train_df)} train images in {public_path}/train, but got {len(list((public_path / 'train').glob('**/*.png')))}"
    assert len(list((public_path / "test").glob("**/*.png"))) == len(
        test_df
    ), f"Expected {len(test_df)} test images in {public_path}/test, but got {len(list((public_path / 'test').glob('**/*.png')))}"

    assert "image_id" in sample_submission.columns, "Sample submission must have 'image_id' column"
    assert "InChI" in sample_submission.columns, "Sample submission must have 'InChI' column"
    assert len(sample_submission) == len(
        test_df
    ), f"Expected {len(test_df)} images in sample submission, but got {len(sample_submission)}"


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split in parallel public_val and private_val directories.
    """
    # Load train data
    old_train = pd.read_csv(raw / "train_labels.csv")

    # ---- 1. Create the original train/test split ----
    # This split is for the main competition test set.
    new_train, new_test = train_test_split(old_train, test_size=0.2, random_state=0)

    # Generate the files for the original public and private directories
    _create_split_files(
        train_df=new_train,
        test_df=new_test,
        raw_images_path=raw / "train",
        public_path=public,
        private_path=private,
    )

    # ---- 2. Create the new validation split ----
    # This second split is performed on the `new_train` set created above.
    # We want the new `test_val` to be the same size as the original `new_test`.
    # Original test size = 0.2 * total. Original train size = 0.8 * total.
    # New test size relative to train set = 0.2 / 0.8 = 0.25
    train_val, test_val = train_test_split(new_train, test_size=0.25, random_state=0)

    # Define the new parallel directories for the validation set
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Generate the files for the new validation directories
    _create_split_files(
        train_df=train_val,
        test_df=test_val,
        raw_images_path=raw / "train",
        public_path=public_val,
        private_path=private_val,
    )

    # ---- 3. Copy shared files and run final checks ----

    # Copy other files into both public and public_val directories
    shutil.copyfile(src=raw / "extra_approved_InChIs.csv", dst=public / "extra_approved_InChIs.csv")
    shutil.copyfile(src=raw / "extra_approved_InChIs.csv", dst=public_val / "extra_approved_InChIs.csv")

    # Original split checks
    assert len(new_train) + len(new_test) == len(
        old_train
    ), f"Expected {len(old_train)} total images in new_train ({len(new_train)}) and new_test ({len(new_test)})"

    # New validation split checks
    assert len(train_val) + len(test_val) == len(
        new_train
    ), f"Expected {len(new_train)} total images in train_val ({len(train_val)}) and test_val ({len(test_val)})"
    # Ensure the size of the validation test set is approx. the same as the original test set
    assert abs(len(test_val) - len(new_test)) <= 1, "Validation test set size should match original test set size"


"""
Core modules for DSLighting simplified API.
"""

from dslighting.core.agent import Agent, AgentResult
from dslighting.core.data_loader import DataLoader, LoadedData

__all__ = [
    "Agent",
    "AgentResult",
    "DataLoader",
    "LoadedData",
]



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import read_csv


def _create_and_populate_split(
    input_df: pd.DataFrame,
    test_size: float,
    random_state: int,
    raw_data_path: Path,
    public_dest_path: Path,
    private_dest_path: Path,
) -> (pd.DataFrame, pd.DataFrame):
    """
    Splits a dataframe and populates the corresponding public and private directories.

    This function performs a train-test split, copies the necessary .npy files,
    creates the ground truth file for the test set, and a sample submission file.
    It's designed to be called multiple times for creating different data splits
    (e.g., train/test and train/validation).

    Args:
        input_df: The dataframe to be split.
        test_size: The proportion of the dataset to allocate to the test split.
        random_state: The seed used by the random number generator.
        raw_data_path: The path to the raw data directory.
        public_dest_path: The destination path for public-facing files.
        private_dest_path: The destination path for private/solution files.

    Returns:
        A tuple containing the training and testing dataframes from the split.
    """
    # Ensure destination directories exist
    public_dest_path.mkdir(parents=True, exist_ok=True)
    private_dest_path.mkdir(parents=True, exist_ok=True)

    # 1. Create train, test from the input dataframe
    split_train, split_test = train_test_split(
        input_df, test_size=test_size, random_state=random_state
    )

    # 2. Copy over files
    split_train.to_csv(public_dest_path / "train_labels.csv", index=False)
    split_test.to_csv(private_dest_path / "test.csv", index=False)

    # Copy shared, non-data-specific files if they exist
    if (raw_data_path / "old_leaky_data").exists():
        shutil.copytree(
            raw_data_path / "old_leaky_data",
            public_dest_path / "old_leaky_data",
            dirs_exist_ok=True,
        )

    for file_id in tqdm(split_train["id"], desc=f"Copying train files to {public_dest_path.name}"):
        subdir = file_id[0]
        src = raw_data_path / "train" / subdir / f"{file_id}.npy"
        dst = public_dest_path / "train" / subdir / f"{file_id}.npy"
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(src, dst)

    for file_id in tqdm(split_test["id"], desc=f"Copying test files to {public_dest_path.name}"):
        subdir = file_id[0]
        src = raw_data_path / "train" / subdir / f"{file_id}.npy"
        dst = public_dest_path / "test" / subdir / f"{file_id}.npy"
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(src, dst)

    # 3. Create sample submission
    sample_submission = split_test.copy()
    sample_submission["target"] = 0.5  # Overwrite with dummy values
    sample_submission.to_csv(public_dest_path / "sample_submission.csv", index=False)

    # 4. Perform checks for this specific split
    assert len(sample_submission) == len(
        split_test
    ), f"Sample submission length mismatch in {public_dest_path.name}."
    assert not set(split_train["id"]).intersection(
        set(split_test["id"])
    ), f"Overlapping IDs in train/test sets for {public_dest_path.name}."

    train_files = {
        file_path.name
        for file_path in (public_dest_path / "train").rglob("*.npy")
        if file_path.is_file()
    }
    test_files = {
        file_path.name
        for file_path in (public_dest_path / "test").rglob("*.npy")
        if file_path.is_file()
    }

    assert len(train_files) == len(
        split_train
    ), f"Train file count mismatch in {public_dest_path.name}."
    assert len(test_files) == len(
        split_test
    ), f"Test file count mismatch in {public_dest_path.name}."
    assert train_files.isdisjoint(
        test_files
    ), f"Overlapping files in train/test directories for {public_dest_path.name}."

    return split_train, split_test


def prepare(raw: Path, public: Path, private: Path):
    """
    Prepares the data by creating two parallel splits:
    1. A standard train/test split (outputs to `public` and `private`).
    2. A subsequent train/validation split (outputs to `public_val` and `private_val`).
    """
    # Read the full raw training data manifest
    full_train_df = read_csv(raw / "train_labels.csv")

    # --- Stage 1: Create the original train/test split ---
    # This split generates the main competition files in `public` and `private`.
    # The logic and outputs here remain identical to the original script.
    print("--- Creating main train/test split ---")
    new_train, new_test = _create_and_populate_split(
        input_df=full_train_df,
        test_size=0.1,
        random_state=0,
        raw_data_path=raw,
        public_dest_path=public,
        private_dest_path=private,
    )
    print("--- Main train/test split created successfully ---\n")

    # --- Stage 2: Create the new train/validation split ---
    # This split takes the `new_train` set from Stage 1 and splits it again.
    # The outputs go into new, parallel directories (`public_val`, `private_val`).
    # The number of samples in the validation set (`test_val`) is designed to be
    # the same as the number of samples in the original test set (`new_test`).
    print("--- Creating train/validation split ---")
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # To make the new validation set size equal to the original test set size,
    # we calculate the required proportion.
    # test_size = len(new_test) / len(new_train)
    # Since len(new_test) is 0.1 * total and len(new_train) is 0.9 * total,
    # the ratio is 0.1 / 0.9 = 1/9.
    validation_test_size = 1 / 9

    # We reuse the same function, ensuring the logic is identical.
    # The returned dataframes are not needed, as this is the final split.
    _create_and_populate_split(
        input_df=new_train,  # Use the training set from the first split as input
        test_size=validation_test_size,
        random_state=0,  # Use the same random state for consistency
        raw_data_path=raw,
        public_dest_path=public_val,
        private_dest_path=private_val,
    )
    print("--- Train/validation split created successfully ---")


from pathlib import Path

import pandas as pd

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path) -> None:
    """Prepare Reverse Game of Life tables for MLE-Bench."""
    public.mkdir(parents=True, exist_ok=True)
    private.mkdir(parents=True, exist_ok=True)

    train = read_csv(raw / "train.csv")
    test = read_csv(raw / "test.csv")
    answers = read_csv(raw / "test_answer.csv")

    sample_path = raw / "sample_submission.csv"
    if sample_path.exists():
        sample_submission = read_csv(sample_path)
    else:
        start_cols = [col for col in answers.columns if col.startswith("start_")]
        sample_submission = answers[["id", *start_cols]].copy()
        sample_submission[start_cols] = 0

    merged_private = test.merge(answers, on="id", how="left", validate="one_to_one")
    start_cols = [col for col in answers.columns if col.startswith("start_")]
    if merged_private[start_cols].isna().any().any():
        missing_ids = merged_private.loc[merged_private[start_cols].isna().any(axis=1), "id"].tolist()
        raise ValueError(f"Missing start cells for ids: {missing_ids[:5]}")

    train.to_csv(public / "train.csv", index=False)
    test.to_csv(public / "test.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    merged_private.to_csv(private / "test.csv", index=False)
    answers.to_csv(private / "gold_submission.csv", index=False)

    assert len(test) == len(answers) == len(merged_private), "Public test and answers row counts differ."
    assert sample_submission.columns.tolist()[0] == "id", "Sample submission must start with 'id'."
    for col in start_cols:
        assert col in sample_submission.columns, f"Sample submission missing column {col}."



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import extract, get_logger, read_csv

logger = get_logger(__name__)


def _create_split(
    input_df: pd.DataFrame,
    test_size: float,
    random_state: int,
    raw_images_path: Path,
    public_path: Path,
    private_path: Path,
    dev_mode: bool = False,
) -> pd.DataFrame:
    """
    Helper function to perform a data split, create necessary files, and organize directories.

    Args:
        input_df: The dataframe to be split.
        test_size: The proportion of the dataset to allocate to the test split.
        random_state: The seed used by the random number generator.
        raw_images_path: Path to the directory containing all source images.
        public_path: The target public directory.
        private_path: The target private directory.
        dev_mode: If True, uses a small sample for faster processing.

    Returns:
        The training portion of the split dataframe.
    """
    # Create train, test from the input dataframe
    locations = input_df["location"].unique()
    train_locations, test_locations = train_test_split(
        locations, test_size=test_size, random_state=random_state
    )

    input_df["split"] = input_df["location"].apply(
        lambda loc: "test" if loc in test_locations else "train"
    )

    train_df = input_df[input_df["split"] == "train"].drop(columns=["split"])
    answers_df = input_df[input_df["split"] == "test"].drop(columns=["split"])

    logger.debug("Train locations: %s", train_locations)
    logger.debug("Test locations: %s", test_locations)
    logger.debug(
        "Test size for this split: %s",
        len(answers_df) / (len(train_df) + len(answers_df)),
    )

    input_df.drop(columns=["split"], inplace=True)  # Drop helper column

    test_df = answers_df.copy().drop(columns=["category_id"])
    gold_submission_df = answers_df.copy()[["id", "category_id"]]
    gold_submission_df.rename(
        columns={"id": "Id", "category_id": "Category"}, inplace=True
    )

    # Make sample submission
    submission_df = test_df.copy()[["id"]]
    submission_df["category_id"] = 0
    submission_df.rename(columns={"id": "Id", "category_id": "Category"}, inplace=True)

    # Checks
    assert set(train_df["id"]).isdisjoint(
        set(test_df["id"])
    ), "train_df and test_df are not disjoint"
    assert len(train_df) + len(test_df) == len(
        input_df
    ), "Length of train_df and test_df should be equal to the length of the input dataframe"
    assert len(answers_df) == len(
        test_df
    ), "Length of answers_df should be equal to the length of test_df"
    assert len(submission_df) == len(
        answers_df
    ), "Length of answers_df should be equal to the length of the sample submission"
    assert (
        input_df.columns.tolist() == train_df.columns.tolist()
    ), f"train_df should have the same columns as the input dataframe: input_df: {input_df.columns.tolist()} != train_df: {train_df.columns.tolist()}"
    assert set(train_df["location"]).isdisjoint(
        set(test_df["location"])
    ), "train_df and test_df should not share any locations"

    # Create directories
    public_path.mkdir(exist_ok=True, parents=True)
    private_path.mkdir(exist_ok=True, parents=True)

    # Write CSVs
    answers_df.to_csv(private_path / "test.csv", index=False)
    gold_submission_df.to_csv(private_path / "answers.csv", index=False)
    train_df.to_csv(public_path / "train.csv", index=False)
    test_df.to_csv(public_path / "test.csv", index=False)
    submission_df.to_csv(public_path / "sample_submission.csv", index=True)

    # Prepare for file copy
    public_train_images = public_path / "train_images"
    public_test_images = public_path / "test_images"
    public_train_images.mkdir(exist_ok=True)
    public_test_images.mkdir(exist_ok=True)

    loop_train_df = train_df.sample(n=100) if dev_mode else train_df
    loop_test_df = test_df.sample(n=100) if dev_mode else test_df

    for file_id in tqdm(loop_train_df["id"], desc=f"Copying train images to {public_path.name}"):
        shutil.copyfile(
            src=raw_images_path / f"{file_id}.jpg",
            dst=public_train_images / f"{file_id}.jpg",
        )

    for file_id in tqdm(loop_test_df["id"], desc=f"Copying test images to {public_path.name}"):
        shutil.copyfile(
            src=raw_images_path / f"{file_id}.jpg",
            dst=public_test_images / f"{file_id}.jpg",
        )

    # Check integrity of the files copied
    assert len(list(public_test_images.glob("*.jpg"))) == len(
        loop_test_df["id"].unique()
    ), f"Public test images in {public_path.name} should have the same number of images as the unique ids in the test set"
    assert len(list(public_train_images.glob("*.jpg"))) == len(
        loop_train_df["id"].unique()
    ), f"Public train images in {public_path.name} should have the same number of images as the unique ids in the train set"

    # Zip up image directories and delete non-zipped files
    shutil.make_archive(
        public_path / "train_images", "zip", public_train_images
    )
    shutil.make_archive(public_path / "test_images", "zip", public_test_images)
    shutil.rmtree(public_train_images)
    shutil.rmtree(public_test_images)

    return train_df


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split in public_val/private_val directories.
    """
    dev_mode = False
    test_size = 0.1
    random_state = 8  # We target a 44% test set size, we have empirically trialed seeds and landed on 8 to achieve this

    # --- Setup and Initial Data Extraction ---
    old_train = read_csv(raw / "train.csv")
    raw_images_path = raw / "train_images"
    raw_images_path.mkdir(exist_ok=True)
    logger.info("Extracting raw images...")
    extract(raw / "train_images.zip", raw_images_path)
    assert len(list(raw_images_path.glob("*.jpg"))) == len(
        old_train["id"].unique()
    ), f"Raw train images should have the same number of images as the unique ids in the old train set, but got {len(list(raw_images_path.glob('*.jpg')))} files and {len(old_train['id'].unique())} ids"

    # --- First Split (Original Public/Private) ---
    # This creates the main competition data. The outputs in `public` and `private`
    # will be identical to the original script's output.
    logger.info("Creating original train/test split for competition...")
    train_from_first_split = _create_split(
        input_df=old_train,
        test_size=test_size,
        random_state=random_state,
        raw_images_path=raw_images_path,
        public_path=public,
        private_path=private,
        dev_mode=dev_mode,
    )

    # --- Second Split (New Validation Set) ---
    # This takes the training data from the first split and splits it again
    # to create a new, smaller training set and a validation set.
    # The output structure in `public_val` and `private_val` mirrors the original.
    logger.info("Creating validation train/test split for local development...")
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"
    _create_split(
        input_df=train_from_first_split,
        test_size=test_size,
        random_state=random_state,  # Use same params to replicate splitting logic
        raw_images_path=raw_images_path,
        public_path=public_val,
        private_path=private_val,
        dev_mode=dev_mode,
    )

    # Clean up the extracted raw images directory
    logger.info("Cleaning up extracted raw images...")
    shutil.rmtree(raw_images_path)
    logger.info("Data preparation complete.")


import json
import shutil
from pathlib import Path

import pandas as pd
from tqdm.auto import tqdm

from mlebench.utils import read_csv


# Create train, test from train split
def create_train_test_split(train_ancestors_df: pd.DataFrame, test_size: int):
    """
    On Kaggle, a user may "fork" (that is, copy) the notebook of another user to create their own version.
    `train_ancestors_df` contains information about the ancestry of each notebook in the training set.

    To create the test split, we select rows from the train set that don't share ancestors with any others
    so that there aren't any relatives in the train set. The train split is the remaining rows.
    """
    group_by_ancestor = {}
    for _, row in tqdm(
        train_ancestors_df.iterrows(), desc="Grouping by Ancestor", total=len(train_ancestors_df)
    ):
        ancestor_id = row["ancestor_id"]
        if ancestor_id not in group_by_ancestor:
            group_by_ancestor[ancestor_id] = []
        group_by_ancestor[ancestor_id].append(row)

    train_ids = []
    test_ids = []
    num_test_ids = 0
    for ancestor_id, rows in tqdm(group_by_ancestor.items(), desc="Splitting on Ancestors"):
        if num_test_ids + len(rows) <= test_size:
            test_ids.extend(([row["id"] for row in rows]))
            num_test_ids += len(rows)
        else:
            train_ids.extend([row["id"] for row in rows])

    assert len(test_ids) == test_size
    assert len(train_ids) == len(train_ancestors_df) - test_size

    return train_ids, test_ids


def prepare(raw: Path, public: Path, private: Path):
    train_ancestors_df = read_csv(raw / "train_ancestors.csv")
    # Shuffle the train_ancestors_df to ensure our split is random
    train_ancestors_df = train_ancestors_df.sample(frac=1, random_state=0).reset_index(drop=True)
    new_train_ids, new_test_ids = create_train_test_split(train_ancestors_df, test_size=20000)

    # Copy json files to public
    (public / "train").mkdir(parents=True, exist_ok=True)
    for train_id in tqdm(new_train_ids, desc="Copying train json files"):
        shutil.copy(raw / "train" / f"{train_id}.json", public / "train" / f"{train_id}.json")
    (public / "test").mkdir(parents=True, exist_ok=True)
    for test_id in tqdm(new_test_ids, desc="Copying test json files"):
        shutil.copy(raw / "train" / f"{test_id}.json", public / "test" / f"{test_id}.json")

    # Generate answers for train and test
    train_orders = read_csv(raw / "train_orders.csv")
    # Answers for new train
    train_orders_new = train_orders[train_orders["id"].isin(new_train_ids)]
    train_orders_new.to_csv(public / "train_orders.csv", index=False)
    # Answers for new test
    test_orders_new = train_orders[train_orders["id"].isin(new_test_ids)]
    test_orders_new.to_csv(private / "test_orders.csv", index=False)

    # Make new train_ancestors.csv, excluding the new_test_ids
    train_ancestors_df = train_ancestors_df[~train_ancestors_df["id"].isin(new_test_ids)]
    train_ancestors_df.to_csv(public / "train_ancestors.csv", index=False)

    # Create sample submission (use the given order without changing it)
    sample_submission_rows = []
    for sample_id in tqdm(test_orders_new["id"], desc="Creating sample submission"):
        # Get cell order from json file
        with open(public / "test" / f"{sample_id}.json") as f:
            json_data = json.load(f)
            cell_order = list(json_data["cell_type"].keys())
        sample_submission_rows.append({"id": sample_id, "cell_order": " ".join(cell_order)})
    sample_submission = pd.DataFrame(sample_submission_rows)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)
    assert len(sample_submission) == len(
        new_test_ids
    ), "Sample submission length does not match test length."



from pathlib import Path
import pandas as pd

from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def _split_and_save(
    source_df: pd.DataFrame,
    public_dir: Path,
    private_dir: Path,
    test_size: float,
    random_state: int,
) -> pd.DataFrame:
    """
    Splits a dataframe into train and test sets and saves them to the specified
    public and private directories, mimicking the competition data format.

    Args:
        source_df: The dataframe to be split.
        public_dir: The directory for public-facing files (train set, unlabeled test set).
        private_dir: The directory for private-facing files (test set labels).
        test_size: The proportion of the dataset to allocate to the test split.
        random_state: The seed used by the random number generator.

    Returns:
        The training set dataframe resulting from the split.
    """
    # Create train, test split
    new_train, new_test = train_test_split(
        source_df, test_size=test_size, random_state=random_state
    )
    new_test_without_labels = new_test.drop(columns=["answer_start", "answer_text"])

    # make private test match submission format
    new_test_labels = new_test[["id", "answer_text"]]
    new_test_labels.columns = ["id", "PredictionString"]

    # Ensure output directories exist
    public_dir.mkdir(exist_ok=True)
    private_dir.mkdir(exist_ok=True)
    
    # Copy over files
    new_train.to_csv(public_dir / "train.csv", index=False)
    new_test_without_labels.to_csv(public_dir / "test.csv", index=False)
    new_test_labels.to_csv(private_dir / "test.csv", index=False)

    # Create sample submission
    sample_submission = new_test_labels.copy()
    sample_submission["PredictionString"] = "dummy text"
    sample_submission.to_csv(public_dir / "sample_submission.csv", index=False)

    assert len(sample_submission) == len(
        new_test_labels
    ), "Sample submission length does not match test length."

    return new_train


def prepare(raw: Path, public: Path, private: Path):

    # --- Define paths for the new validation split ---
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Load the initial raw training data
    original_train_df = read_csv(raw / "train.csv")

    # --- First Split: Create original train/test sets ---
    # This split generates the main competition files in `public` and `private`.
    # The outputs of this step must remain identical to the original script.
    main_train_set = _split_and_save(
        source_df=original_train_df,
        public_dir=public,
        private_dir=private,
        test_size=0.1,
        random_state=0,
    )

    # --- Second Split: Create validation train/test sets from the main train set ---
    # This split uses the *training data* from the first split as its source.
    # The new `test_val` set will have approx. the same size as the original `test` set.
    # test_size for 2nd split = (size of original test) / (size of new train)
    # = (0.1 * total) / (0.9 * total) = 0.1 / 0.9
    validation_test_size = 0.1 / (1.0 - 0.1)

    _split_and_save(
        source_df=main_train_set,
        public_dir=public_val,
        private_dir=private_val,
        test_size=validation_test_size,
        random_state=0,  # Use the same random state for consistency
    )


import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def _save_split(
    train_df: pd.DataFrame, test_df: pd.DataFrame, public_path: Path, private_path: Path
):
    """
    Saves a given train/test split to the specified public and private directories.

    This helper function ensures a consistent file structure and naming convention
    for any data split.

    Args:
        train_df: The training dataframe.
        test_df: The testing dataframe (with labels).
        public_path: The path to the public output directory.
        private_path: The path to the private output directory.
    """
    # Ensure output directories exist
    public_path.mkdir(parents=True, exist_ok=True)
    private_path.mkdir(parents=True, exist_ok=True)

    # Prepare dataframes for saving
    test_without_labels = test_df.drop(columns=["fare_amount"])
    submission_df = test_df.copy()[["key"]]
    submission_df["fare_amount"] = 11.35  # A sample constant value

    # Write CSVs to public and private directories
    train_df.to_csv(public_path / "train.csv", index=False)
    test_without_labels.to_csv(public_path / "test.csv", index=False)
    submission_df.to_csv(public_path / "sample_submission.csv", index=False)
    test_df.to_csv(private_path / "test.csv", index=False)

    # Run checks to ensure data integrity
    assert set(train_df["key"]).isdisjoint(
        set(test_df["key"])
    ), "Train and test sets share samples!"
    assert test_df.shape[1] == 8, f"Test set should have 8 columns, but has {test_df.shape[1]}"
    assert (
        test_without_labels.shape[1] == 7
    ), f"Test set without labels should have 7 columns, but has {test_without_labels.shape[1]}"
    assert train_df.shape[1] == 8, f"Train set should have 8 columns, but has {train_df.shape[1]}"
    assert (
        submission_df.shape[1] == 2
    ), f"Sample submission should have 2 columns, but has {submission_df.shape[1]}"
    assert (
        submission_df.shape[0] == test_df.shape[0]
    ), f"Sample submission should have {test_df.shape[0]} rows, but has {submission_df.shape[0]}"


def prepare(raw: Path, public: Path, private: Path):
    # Load the raw training data
    old_train = read_csv(raw / "train.csv")

    # --- 1. Create the original train/test split for the main competition files ---
    # This split creates the primary train and test sets.
    # The outputs in `public/` and `private/` will not be changed.
    train_orig, test_orig = train_test_split(old_train, test_size=9914, random_state=0)

    # Save the original split results
    _save_split(
        train_df=train_orig, test_df=test_orig, public_path=public, private_path=private
    )

    # Copy over other files to the original public directory
    shutil.copy(raw / "GCP-Coupons-Instructions.rtf", public / "GCP-Coupons-Instructions.rtf")

    # --- 2. Create the new train/validation split for local validation ---
    # This second split takes the `train_orig` set and splits it again to create
    # a smaller training set and a new validation set (`test_val`).
    # The logic and test size are identical to the first split.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    train_val, test_val = train_test_split(train_orig, test_size=9914, random_state=0)

    # Save the validation split results to the new directories
    _save_split(
        train_df=train_val, test_df=test_val, public_path=public_val, private_path=private_val
    )

    # 复制相同的文件到 public_val 目录
    shutil.copy(raw / "GCP-Coupons-Instructions.rtf", public_val / "GCP-Coupons-Instructions.rtf")


"""
Implements DSAgentState, which manages the running log for the DS-Agent workflow.
"""
from pydantic import BaseModel
from dsat.services.states.base import State

class DSAgentState(State, BaseModel):
    """
    Holds the state for a DS-Agent workflow execution, primarily the running log
    which accumulates summaries of each experimental step.
    """
    running_log: str = ""
    final_code: str = ""



# dsat/runner.py
import logging
import shutil
import uuid
import json
import time
from datetime import datetime
from pathlib import Path
from typing import Callable, Coroutine, Any, Tuple, Dict, Type, Optional, List

# Core configuration and models
from dsat.config import DSATConfig
from dsat.models.task import TaskDefinition, TaskType

# Services and workflows
from dsat.services.llm import LLMService
from dsat.workflows.base import DSATWorkflow

# Dynamic components (factories and handlers)
from dsat.tasks.handlers import TaskHandler, KaggleTaskHandler, QATaskHandler, DataSciTaskHandler, OpenEndedTaskHandler
from dsat.workflows.factory import (
    WorkflowFactory,
    AutoMindWorkflowFactory,
    AIDEWorkflowFactory,
    DSAgentWorkflowFactory,
    DataInterpreterWorkflowFactory,
    AutoKaggleWorkflowFactory,
    AFlowWorkflowFactory,
    DeepAnalyzeWorkflowFactory,
    DynamicWorkflowFactory,
)
# Import AFlow workflow for type checking
from dsat.workflows.search.aflow_workflow import AFlowWorkflow
from dsat.services.states.journal import JournalState

logger = logging.getLogger(__name__)

# ==============================================================================
# ==                            COMPONENT REGISTRIES                          ==
# ==============================================================================

WORKFLOW_FACTORIES: Dict[str, WorkflowFactory] = {
    "automind": AutoMindWorkflowFactory(),
    "aide": AIDEWorkflowFactory(),
    "dsagent": DSAgentWorkflowFactory(),
    "data_interpreter": DataInterpreterWorkflowFactory(),
    "autokaggle": AutoKaggleWorkflowFactory(),
    "aflow": AFlowWorkflowFactory(),
    "deepanalyze": DeepAnalyzeWorkflowFactory(),
}

TASK_HANDLER_CLASSES: Dict[TaskType, Type[TaskHandler]] = {
    "kaggle": KaggleTaskHandler,
    "qa": QATaskHandler,
    "datasci": DataSciTaskHandler,
    "open_ended": OpenEndedTaskHandler,
    # "code": CodeTaskHandler, # future extension
}


# ==============================================================================
# ==                                DSAT RUNNER                               ==
# ==============================================================================

class DSATRunner:
    """
    Orchestrates benchmarking tasks by instantiating workflows, preparing inputs,
    executing runs, and collecting telemetry for later inspection.
    """

    def __init__(self, config: DSATConfig):
        logger.info(f"Initializing DSATRunner for workflow: '{config.workflow.name}'")
        self.config = config
        self.factories = WORKFLOW_FACTORIES.copy()
        self.factory = self.factories.get(config.workflow.name)
        if not self.factory:
            available = ", ".join(self.factories.keys())
            raise ValueError(f"Unknown workflow '{config.workflow.name}'. Available workflows: [{available}]")

        self.handler_classes = TASK_HANDLER_CLASSES
        self.benchmark = None
        self.run_records: List[Dict[str, Any]] = []

        logger.info("DSATRunner is ready to evaluate tasks.")

    def register_workflow(self, name: str, factory: WorkflowFactory) -> None:
        """
        Register a workflow factory dynamically for this runner instance.
        Critical for paradigms like AFLOW which synthesize workflows at runtime.
        """
        logger.info(f"Registering workflow '{name}' for this runner instance.")
        self.factories[name] = factory
        if self.config.workflow and self.config.workflow.name == name:
            self.factory = factory
            logger.info(f"Active workflow factory switched to '{name}'.")

    def get_eval_function(self) -> Callable[[TaskDefinition], Coroutine[Any, Any, Tuple[Any, float, Dict[str, Any]]]]:
        """
        Produce an async function that evaluates a single TaskDefinition and returns (result, cost, usage_summary).
        Benchmark drivers call this function repeatedly for each competition/task.
        """

        async def eval_function(task: TaskDefinition) -> Tuple[Any, float, Dict[str, Any]]:
            logger.info(f"Starting evaluation for task '{task.task_id}' (type='{task.task_type}').")

            # If a specific run name is provided in config, use it. Otherwise, generate one.
            if self.config.run.name and self.config.run.name != "dsat_run":
                task_run_name = self.config.run.name
            else:
                safe_task_id = "".join(c if c.isalnum() else "_" for c in task.task_id)
                unique_suffix = uuid.uuid4().hex[:8]
                task_run_name = f"{self.config.run.name}_{safe_task_id}_{unique_suffix}"

            task_config = self.config.model_copy(deep=True)
            task_config.run.name = task_run_name

            workflow: Optional[DSATWorkflow] = None
            workspace_service = None
            sandbox_service = None
            llm_service: Optional[LLMService] = None
            result: Any = None
            run_total_cost = 0.0
            run_started_at = datetime.utcnow()
            run_perf_start = time.perf_counter()

            benchmark_instance = self.benchmark

            try:
                workflow = self.factory.create_workflow(task_config, benchmark=benchmark_instance)
                workspace_service = workflow.services.get("workspace")
                llm_service = workflow.services.get("llm")
                sandbox_service = workflow.services.get("sandbox")

                if isinstance(workflow, AFlowWorkflow):
                    optimizer_name = "AFLOW"
                    logger.info("Detected %s workflow. Running meta-optimization stage.", optimizer_name)
                    best_workflow_code = await workflow.optimize()
                    logger.info("Meta-optimization complete. Proceeding with final evaluation workflow.")

                    if hasattr(benchmark_instance, 'set_mode'):
                        logger.info("Switching benchmark to 'test' mode for final %s evaluation.", optimizer_name)
                        benchmark_instance.set_mode('test')

                    dynamic_factory = DynamicWorkflowFactory(code_string=best_workflow_code)
                    workflow = dynamic_factory.create_workflow(task_config, benchmark=benchmark_instance)
                    llm_service = workflow.services.get("llm")
                    sandbox_service = workflow.services.get("sandbox")
                    workspace_service = workflow.services.get("workspace")
                    logger.info("Final %s workflow instantiated and ready.", optimizer_name)

                workspace_service = workspace_service or workflow.services.get("workspace")
                llm_service = llm_service or workflow.services.get("llm")
                sandbox_service = sandbox_service or workflow.services.get("sandbox")

                if not llm_service:
                    logger.error("Workflow did not expose an LLMService.")
                    return "[ERROR] Missing LLM service", 0.0

                handler_class = self.handler_classes.get(task.task_type)
                if not handler_class:
                    logger.error(f"No handler registered for task type '{task.task_type}'.")
                    return f"[ERROR] Unsupported task type '{task.task_type}'", 0.0

                handler: TaskHandler = handler_class()

                description, io_instructions = "", ""
                data_dir, output_path = None, None

                try:
                    description, io_instructions, data_dir, output_path = handler.prepare_input(task)

                    if workspace_service:
                        try:
                            workspace_service.link_data_to_workspace(data_dir)
                        except Exception as link_error:
                            raise RuntimeError(f"Failed to link data directory: {link_error}") from link_error
                    else:
                        logger.warning("WorkspaceService missing; skipping data linkage.")

                    await workflow.solve(
                        description=description,
                        io_instructions=io_instructions,
                        data_dir=data_dir,
                        output_path=output_path
                    )

                    if workspace_service and output_path:
                        sandbox_workdir = workspace_service.get_path("sandbox_workdir")
                        generated_file = sandbox_workdir / output_path.name
                        if generated_file.exists():
                            # Only copy if output_path is an absolute path
                            # If it's a relative path (just filename), keep it in workspace/sandbox
                            if output_path.is_absolute():
                                output_path.parent.mkdir(parents=True, exist_ok=True)
                                if generated_file.resolve() != output_path.resolve():
                                    logger.info(f"Collecting produced artifact '{output_path.name}' from the sandbox.")

                                    # Handle both files and directories (e.g., for open-ended tasks)
                                    if generated_file.is_dir():
                                        # For directories (like 'artifacts'), use copytree
                                        if output_path.exists():
                                            if output_path.is_dir():
                                                shutil.rmtree(output_path)
                                            else:
                                                output_path.unlink()
                                        shutil.copytree(generated_file, output_path)
                                        logger.info(f"Copied directory '{generated_file}' to '{output_path}'")
                                    else:
                                        # For files, use regular copy
                                        shutil.copy(generated_file, output_path)
                                        logger.info(f"Copied file '{generated_file}' to '{output_path}'")
                            else:
                                # Relative path - file stays in sandbox, use workspace path for result
                                logger.info(f"Output file '{output_path.name}' remains in workspace sandbox: {generated_file}")
                                # Update output_path to point to the actual file in workspace
                                output_path = generated_file
                        else:
                            logger.warning(f"No output '{output_path.name}' found in sandbox '{sandbox_workdir}' after workflow execution.")

                    if output_path:
                        result = handler.parse_output(output_path)

                        # Grade the submission if benchmark is available
                        if benchmark_instance and hasattr(benchmark_instance, 'grade') and isinstance(result, Path):
                            try:
                                logger.info(f"Grading submission: {result}")
                                score = await benchmark_instance.grade(result)
                                logger.info(f"✓ Grading complete | Score: {score}")
                                # Return score as result
                                result = {"score": score, "submission_path": str(result)}
                            except Exception as grade_error:
                                logger.warning(f"Grading failed: {grade_error}")
                                # Keep the path as result if grading fails
                                logger.info(f"Submission created at: {result}")
                        elif isinstance(result, Path):
                            logger.info(f"Submission created at: {result}")

                    logger.info(f"Task '{task.task_id}' evaluation finished successfully.")

                except Exception as execution_error:
                    logger.error(f"Task '{task.task_id}' failed: {execution_error}", exc_info=True)
                    result = f"[ERROR] {execution_error.__class__.__name__}: {execution_error}"
                finally:
                    handler.cleanup()
                    if workspace_service:
                        ended_at = datetime.utcnow()
                        duration_sec = round(time.perf_counter() - run_perf_start, 4)
                        run_total_cost = llm_service.get_total_cost() if llm_service else 0.0
                        try:
                            self._persist_run_metadata(
                                workspace_service=workspace_service,
                                task_config=task_config,
                                task=task,
                                description=description,
                                io_instructions=io_instructions,
                                data_dir=data_dir,
                                output_path=output_path,
                                result=result,
                                llm_service=llm_service,
                                sandbox_service=sandbox_service,
                                workflow=workflow,
                                started_at=run_started_at,
                                ended_at=ended_at,
                                duration_seconds=duration_sec,
                                total_cost=run_total_cost
                            )
                        except Exception as persist_error:
                            logger.error(f"Failed to persist telemetry for task '{task.task_id}': {persist_error}", exc_info=True)

                        failed = isinstance(result, str) and result.startswith("[ERROR]")
                        keep_on_fail = self.config.run.keep_workspace_on_failure
                        keep_all = self.config.run.keep_all_workspaces
                        workspace_service.cleanup(keep_workspace=keep_all or (failed and keep_on_fail))

            except Exception as workflow_error:
                logger.error(f"Workflow creation failed for task '{task.task_id}': {workflow_error}", exc_info=True)
                result = f"[ERROR] {workflow_error.__class__.__name__}: {workflow_error}"
                if workspace_service:
                    ended_at = datetime.utcnow()
                    duration_sec = round(time.perf_counter() - run_perf_start, 4)
                    run_total_cost = llm_service.get_total_cost() if llm_service else 0.0
                    try:
                        self._persist_run_metadata(
                            workspace_service=workspace_service,
                            task_config=task_config,
                            task=task,
                            description="",
                            io_instructions="",
                            data_dir=None,
                            output_path=None,
                            result=result,
                            llm_service=llm_service,
                            sandbox_service=sandbox_service,
                            workflow=workflow,
                            started_at=run_started_at,
                            ended_at=ended_at,
                            duration_seconds=duration_sec,
                            total_cost=run_total_cost
                        )
                    except Exception as persist_error:
                        logger.error(f"Telemetry persistence failed after workflow creation error: {persist_error}", exc_info=True)
                    workspace_service.cleanup(keep_workspace=True)

            run_total_cost = llm_service.get_total_cost() if llm_service else run_total_cost
            usage_summary = llm_service.get_usage_summary() if llm_service else {}
            logger.info(f"Task '{task.task_id}' LLM cost: ${run_total_cost:.6f}")
            return result, run_total_cost, usage_summary

        return eval_function

    def get_run_records(self) -> List[Dict[str, Any]]:
        """
        Return a shallow copy of stored run metadata records for summary rendering.
        """
        return [record.copy() for record in self.run_records]

    def _persist_run_metadata(
        self,
        *,
        workspace_service,
        task_config: DSATConfig,
        task: TaskDefinition,
        description: str,
        io_instructions: str,
        data_dir: Optional[Path],
        output_path: Optional[Path],
        result: Any,
        llm_service: Optional[LLMService],
        sandbox_service: Optional[Any],
        workflow: Optional[DSATWorkflow],
        started_at: datetime,
        ended_at: datetime,
        duration_seconds: float,
        total_cost: float,
    ) -> None:
        """
        Write per-task telemetry (LLM calls, sandbox runs, search tree, summary) to the workspace.
        """
        telemetry_dir = "telemetry"
        workspace_dir = workspace_service.get_path("run_dir")
        llm_calls = llm_service.get_call_history() if llm_service else []
        sandbox_runs = sandbox_service.get_execution_history() if sandbox_service else []
        usage_summary = llm_service.get_usage_summary() if llm_service else {}
        best_node = self._get_best_node(workflow)
        search_tree_data, search_tree_info = self._extract_search_tree(workflow, best_node)

        config_snapshot = task_config.model_dump()
        if "llm" in config_snapshot and "api_key" in config_snapshot["llm"]:
            config_snapshot["llm"]["api_key"] = "***REDACTED***"

        benchmark_snapshot = self._build_benchmark_snapshot()

        final_code_path: Optional[str] = None
        final_candidate = workspace_service.get_path("artifacts") / "final_submission" / "final_solution.py"
        if final_candidate.exists():
            final_code_path = str(final_candidate)
        elif best_node:
            final_code_path = best_node.final_submission_path or best_node.code_artifact_path

        filtered_parameters = {
            key: value for key, value in (task_config.run.parameters or {}).items()
            if value not in (None, "", [], {})
        }

        metadata = {
            "run_name": task_config.run.name,
            "workspace_dir": str(workspace_dir),
            "workflow": task_config.workflow.name if task_config.workflow else None,
            "parameters": filtered_parameters,
            "benchmark": benchmark_snapshot,
            "task": {
                "task_id": task.task_id,
                "task_type": task.task_type,
                "payload": task.payload,
            },
            "task_context": {
                "description": description,
                "io_instructions": io_instructions,
                "data_dir": str(data_dir) if data_dir else None,
                "expected_output_path": str(output_path) if output_path else None,
            },
            "timeline": {
                "started_at_utc": started_at.isoformat() + "Z",
                "ended_at_utc": ended_at.isoformat() + "Z",
                "duration_seconds": duration_seconds,
            },
            "summary": {
                "result": self._format_result(result),
                "success": not (isinstance(result, str) and result.startswith("[ERROR]")),
                "total_cost": total_cost,
                "usage": usage_summary,
                "cost_per_token": usage_summary.get("cost_per_token"),
                "llm_call_count": len(llm_calls),
                "sandbox_run_count": len(sandbox_runs),
                "final_code": best_node.code if best_node else None,
                "final_code_path": final_code_path,
                "best_node_id": best_node.id if best_node else None,
                "best_path_node_ids": search_tree_info.get("best_path"),
            },
            "config_snapshot": config_snapshot,
        }

        # Save Final Code to a standard location
        if best_node and best_node.code:
            final_code_file = workspace_service.get_path("run_dir") / "final_solution.py"
            with open(final_code_file, "w", encoding="utf-8") as f:
                f.write(best_node.code)
            metadata["summary"]["final_code_path"] = str(final_code_file)

            # 💾 NEW: Save model training code to code_history directory
            try:
                code_history_dir = workspace_service.get_path("sandbox_workdir") / "code_history"
                code_history_dir.mkdir(parents=True, exist_ok=True)

                # Find next available number for model training code
                import re
                existing_model_codes = list(code_history_dir.glob("model_code_*.py"))
                if existing_model_codes:
                    numbers = []
                    for f in existing_model_codes:
                        match = re.search(r'model_code_(\d+)\.py', f.name)
                        if match:
                            numbers.append(int(match.group(1)))
                    next_num = max(numbers) + 1 if numbers else 1
                else:
                    next_num = 1

                # Save with formatted number and metadata
                model_code_filename = f"model_code_{next_num:03d}.py"
                model_code_filepath = code_history_dir / model_code_filename

                # Add header with training metadata
                import datetime
                header = f'''# Code Type: MODEL TRAINING
# Workflow: {task_config.workflow.name if task_config.workflow else 'Unknown'}
# Model: {task_config.llm.model if task_config.llm else 'Unknown'}
# Generated: {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}
# Task ID: {task.task_id}
# Success: {not (isinstance(result, str) and result.startswith("[ERROR]"))}

'''
                model_code_filepath.write_text(header + best_node.code)
                logger.info(f"💾 Saved model training code to workspace: {model_code_filepath}")
            except Exception as e:
                logger.warning(f"Failed to save model training code to code_history: {e}")

        # Save Evaluation Result to a CSV in workspace
        if isinstance(result, (float, int, str)) and not str(result).startswith("[ERROR]"):
            try:
                res_file = workspace_service.get_path("run_dir") / "evaluation_result.csv"
                with open(res_file, "w") as f:
                    f.write("task_id,score,cost,duration\n")
                    f.write(f"{task.task_id},{result},{total_cost},{duration_seconds}\n")
            except Exception:
                pass

        detail_files = {}
        if llm_calls:
            llm_calls_path = f"{telemetry_dir}/llm_calls.jsonl"
            self._write_jsonl(workspace_service, llm_calls_path, llm_calls)
            detail_files["llm_calls"] = f"artifacts/{llm_calls_path}"
        if sandbox_runs:
            sandbox_runs_path = f"{telemetry_dir}/sandbox_runs.jsonl"
            self._write_jsonl(workspace_service, sandbox_runs_path, sandbox_runs)
            detail_files["sandbox_runs"] = f"artifacts/{sandbox_runs_path}"
        if search_tree_data:
            search_tree_path = f"{telemetry_dir}/search_tree.json"
            workspace_service.write_file(
                json.dumps(search_tree_data, ensure_ascii=False, indent=2),
                "artifacts",
                search_tree_path
            )
            detail_files["search_tree"] = f"artifacts/{search_tree_path}"
            metadata["search_tree"] = {
                "node_count": len(search_tree_data),
                "best_node_id": search_tree_info.get("best_node_id"),
                "best_path_node_ids": search_tree_info.get("best_path"),
                "file": f"artifacts/{search_tree_path}",
            }
        else:
            metadata["search_tree"] = None

        if detail_files:
            metadata["detail_files"] = detail_files

        run_metadata_path = f"{telemetry_dir}/run_metadata.json"
        workspace_service.write_file(
            json.dumps(metadata, ensure_ascii=False, indent=2),
            "artifacts",
            run_metadata_path
        )

        metadata_file = workspace_service.get_path("artifacts") / run_metadata_path
        record_entry = {
            "task_id": task.task_id,
            "metadata_path": str(metadata_file),
            "workspace_dir": metadata["workspace_dir"],
            "summary": metadata["summary"],
            "timeline": metadata["timeline"],
            "parameters": metadata["parameters"],
            "detail_files": metadata.get("detail_files"),
        }
        self.run_records.append(record_entry)

    def _write_jsonl(self, workspace_service, relative_path: str, records: List[Dict[str, Any]]) -> None:
        """Write newline-delimited JSON records to an artifacts sub-path."""
        content = "\n".join(json.dumps(record, ensure_ascii=False) for record in records)
        workspace_service.write_file(content, "artifacts", relative_path)

    def _format_result(self, result: Any) -> Any:
        """Return a serialization-friendly representation of the workflow result."""
        if isinstance(result, Path):
            return str(result)
        return result

    def _get_best_node(self, workflow: Optional[DSATWorkflow]):
        if not workflow or not hasattr(workflow, "state"):
            return None
        state = workflow.state
        if isinstance(state, JournalState):
            return state.get_best_node()
        return None

    def _extract_search_tree(self, workflow: Optional[DSATWorkflow], best_node: Optional[Any]):
        if not workflow or not hasattr(workflow, "state"):
            return None, {"best_node_id": None, "best_path": None}
        state = workflow.state
        if not isinstance(state, JournalState):
            return None, {"best_node_id": None, "best_path": None}

        nodes = [
            node.model_dump(mode="json")
            for node in sorted(state.nodes.values(), key=lambda n: n.step)
        ]
        best_path = self._extract_best_path(state, best_node)
        info = {
            "best_node_id": best_node.id if best_node else None,
            "best_path": best_path,
        }
        return nodes, info

    def _extract_best_path(self, state: JournalState, best_node: Optional[Any]) -> Optional[List[str]]:
        if not best_node:
            return None
        path: List[str] = []
        current = best_node
        while current:
            path.append(current.id)
            current = state.get_node(current.parent_id) if current.parent_id else None
        return list(reversed(path))

    def _build_benchmark_snapshot(self) -> Optional[Dict[str, Any]]:
        if not self.benchmark:
            return None
        snapshot: Dict[str, Any] = {"name": getattr(self.benchmark, "name", None)}
        data_dir = getattr(self.benchmark, "data_dir", None)
        if data_dir is not None:
            snapshot["data_dir"] = str(data_dir)
        config_value = getattr(self.benchmark, "config", None)
        if isinstance(config_value, dict):
            snapshot["config"] = config_value
        return snapshot



import shutil
from pathlib import Path
from typing import List

import pandas as pd
from PIL import Image
from tqdm import tqdm

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path) -> None:
    """
    Prepares the data by creating a primary train/test split and a secondary
    train/validation split.

    The primary split uses fragments {1, 2} for training and {3} for testing,
    outputting to `public/` and `private/` directories.

    The secondary split uses fragment {1} for training and {2} for validation,
    outputting to `public_val/` and `private_val/` directories, mirroring the
    structure and logic of the primary split.
    """

    def _create_split_data(
        train_fragment_ids: List[str],
        test_fragment_id: str,
        raw_path: Path,
        public_path: Path,
        private_path: Path,
    ) -> None:
        """
        A generic helper function to process a set of train/test fragments
        and generate the corresponding public and private data.
        """
        # Ensure destination directories exist
        public_path.mkdir(parents=True, exist_ok=True)
        private_path.mkdir(parents=True, exist_ok=True)

        # Copy train images to `public_path/train/{id}/`
        for frag_id in train_fragment_ids:
            shutil.copytree(
                src=raw_path / "train" / frag_id,
                dst=public_path / "train" / frag_id,
                dirs_exist_ok=True,  # Make script re-runnable
            )

        test_fragment_path = raw_path / "train" / test_fragment_id

        # Create test `inklabels_rle.csv`
        inklabels_rle = read_csv(test_fragment_path / "inklabels_rle.csv")

        assert (
            len(inklabels_rle) == 1
        ), f"Expected a single row in `inklabels_rle.csv`, got {len(inklabels_rle)} rows."

        img_path = test_fragment_path / "ir.png"

        assert img_path.is_file(), f"Expected image file at {img_path}, but it does not exist."

        with Image.open(img_path) as img:
            width, height = img.size

        inklabels_rle["width"] = width
        inklabels_rle["height"] = height
        inklabels_rle["Id"] = "a"

        inklabels_rle.to_csv(private_path / "inklabels_rle.csv", index=False)

        # Write `gold_submission.csv`
        inklabels_rle.drop(columns=["width", "height"]).to_csv(
            private_path / "gold_submission.csv",
            index=False,
        )

        # Copy test images to `public_path/test/a/`
        test_imgs = list(test_fragment_path.rglob("*"))

        for fpath in tqdm(test_imgs, desc=f"Creating test images for {public_path.name}"):
            if not fpath.is_file():
                continue

            assert fpath.suffix in [
                ".png",
                ".csv",
                ".tif",
            ], f"Expected file with extension png, csv, or tif, got `{fpath.suffix}` for file `{fpath}`"

            relative_path = fpath.relative_to(test_fragment_path)

            if fpath.name in ["inklabels.png", "inklabels_rle.csv", "ir.png"]:
                continue  # skip test images and labels

            dst = public_path / "test" / "a" / relative_path
            dst.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(fpath, dst)  # everything else to `public_path`

        sample_submission = pd.DataFrame({"Id": ["a"], "Predicted": ["1 1 5 1"]})
        sample_submission.to_csv(public_path / "sample_submission.csv", index=False)

        # Sanity checks
        assert len(sample_submission) == len(inklabels_rle), (
            f"Expected {len(inklabels_rle)} rows in `sample_submission.csv`, got "
            f"{len(sample_submission)} rows."
        )

        actual_sample_submission = read_csv(public_path / "sample_submission.csv")
        actual_inklabels_rle = read_csv(private_path / "inklabels_rle.csv")

        assert (
            "Id" in actual_sample_submission.columns
        ), f"Expected column `Id` in `sample_submission.csv`."
        assert (
            "Predicted" in actual_sample_submission.columns
        ), f"Expected column `Predicted` in `sample_submission.csv`."

        assert "Id" in actual_inklabels_rle.columns, f"Expected column `Id` in `inklabels_rle.csv`."
        assert (
            "Predicted" in actual_inklabels_rle.columns
        ), f"Expected column `Predicted` in `inklabels_rle.csv`."
        assert (
            "width" in actual_inklabels_rle.columns
        ), f"Expected column `width` in `inklabels_rle.csv`."
        assert (
            "height" in actual_inklabels_rle.columns
        ), f"Expected column `height` in `inklabels_rle.csv`."

        for frag_id in train_fragment_ids:
            assert len(list((public_path / "train" / frag_id).rglob("*"))) == len(
                list((raw_path / "train" / frag_id).rglob("*"))
            ), (
                f"Expected {len(list((raw_path / 'train' / frag_id).rglob('*')))} files in `{public_path}/train/{frag_id}`, got "
                f"{len(list((public_path / 'train' / frag_id).rglob('*')))} files."
            )

        n_test_actual = len(list((public_path / "test" / "a").rglob("*")))
        n_test_expected = len(list(test_fragment_path.rglob("*"))) - len(
            ["inklabels.png", "inklabels_rle.csv", "ir.png"]
        )

        assert n_test_actual == n_test_expected, (
            f"Expected " f"{n_test_expected} " f"files in `{public_path}/test/a`, got {n_test_actual} files."
        )

    # --- Primary Split (Original Behavior) ---
    # This call generates the original `public` and `private` directories.
    # The contents will be identical to the original script's output.
    _create_split_data(
        train_fragment_ids=["1", "2"],
        test_fragment_id="3",
        raw_path=raw,
        public_path=public,
        private_path=private,
    )

    # --- New Validation Split ---
    # Define paths for the new validation set directories.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # This call splits the original training data ({1, 2}) into a new, smaller
    # training set ({1}) and a validation set ({2}). The output structure
    # in `public_val` and `private_val` will mirror the original one.
    _create_split_data(
        train_fragment_ids=["1"],
        test_fragment_id="2",
        raw_path=raw,
        public_path=public_val,
        private_path=private_val,
    )


import numpy as np
import pandas as pd

from mlebench.grade_helpers import InvalidSubmissionError

column_order = ["patient_overall", "C1", "C2", "C3", "C4", "C5", "C6", "C7"]


def unmelt_df(melted_df: pd.DataFrame, id_col: str, target_col: str) -> pd.DataFrame:
    unmelted_df = melted_df.copy()
    unmelted_df[["StudyInstanceUID", "variable"]] = unmelted_df[id_col].str.split(
        "_", n=1, expand=True
    )
    unmelted_df = unmelted_df.pivot(
        index="StudyInstanceUID", columns="variable", values=target_col
    ).reset_index()
    return unmelted_df


def binary_log_loss(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    """
    Computes binary log loss for each sample and variable (no reduction).

    Args:
        y_true: True binary labels, shape (n_samples, n_variables)
        y_pred: Predicted probabilities, shape (n_samples, n_variables)

    Returns:
        Binary log loss for each sample and variable, shape (n_samples, n_variables)
    """
    epsilon = 1e-15
    # avoid y_preds of 0 or 1, since we'll be taking the log and that is undefined/inf
    y_pred = np.clip(y_pred, epsilon, 1 - epsilon)
    loss = -(y_true * np.log(y_pred) + (1 - y_true) * np.log(1 - y_pred))
    return loss


def custom_comp_weighted_loss(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Adapted from
    https://www.kaggle.com/competitions/rsna-2022-cervical-spine-fracture-detection/discussion/341854

    Adaptations:
        - custom binary_log_loss function in numpy to avoid torch dependency
        - working in numpy arrays instead of torch tensors
        - explanatory comments
        - added comp weights inside fn
    """
    # https://www.kaggle.com/competitions/rsna-2022-cervical-spine-fracture-detection/discussion/340392
    # assuming ordering [patient_overall, C1, C2, C3, C4, C5, C6, C7]
    competition_weights = {
        # weights for negative labels
        "-": np.array([7, 1, 1, 1, 1, 1, 1, 1], dtype=np.float32),
        # weights for positive labels
        "+": np.array([14, 2, 2, 2, 2, 2, 2, 2], dtype=np.float32),
    }

    # (n_samples, n_variables)
    loss = binary_log_loss(y_true, y_pred)
    # (n_samples, n_variables) weights matrix, conditionally built based on y_true values
    weights = y_true * competition_weights["+"] + (1 - y_true) * competition_weights["-"]
    # (n_samples, n_variables) element-wise multiplication of loss and weights
    loss = loss * weights
    # (n_samples,) sum of weighted losses across variables
    loss = np.sum(loss, axis=1)
    # (n_samples,) normalize the loss by the sum of weights
    loss = loss / weights.sum(axis=1)
    # finally, take the mean across all samples
    return np.mean(loss)


def prepare_for_metric(
    submission: pd.DataFrame, answers: pd.DataFrame
) -> tuple[np.ndarray, np.ndarray]:
    id_col = "row_id"
    target_col = "fractured"

    if id_col not in submission.columns:
        raise InvalidSubmissionError(f"Submission is missing the `{id_col}` column.")
    if target_col not in submission.columns:
        raise InvalidSubmissionError(f"Submission is missing the `{target_col}` column.")
    if not submission[target_col].between(0, 1).all():
        raise InvalidSubmissionError(f"Submission `{target_col}` values must be between 0 and 1.")
    if submission[target_col].isnull().any():
        raise InvalidSubmissionError(f"Submission `{target_col}` values must not be missing.")

    assert id_col in answers.columns, f"Answers are missing the `{id_col}` column."
    assert target_col in answers.columns, f"Answers are missing the `{target_col}` column."

    # all answers should be either 0 or 1
    assert answers[target_col].isin([0, 1]).all(), f"Answers `{target_col}` values must be 0 or 1."
    # all answers should be present
    assert (
        not answers[target_col].isnull().any()
    ), f"Answers `{target_col}` values must not be missing."

    # sort both submission and answers by id_col
    submission = submission.sort_values(by=id_col)
    answers = answers.sort_values(by=id_col)

    # check that the ids match
    if (submission[id_col].values != answers[id_col].values).any():
        raise InvalidSubmissionError("Submission should contain same ids as answers.")

    # checks complete; now we can prepare

    # need to unmelt both submission and answers, back to getting one row per StudyInstanceUID
    unmelted_submission = unmelt_df(melted_df=submission, id_col=id_col, target_col=target_col)
    unmelted_answers = unmelt_df(melted_df=answers, id_col=id_col, target_col=target_col)

    # sort both by StudyInstanceUID
    unmelted_submission = unmelted_submission.sort_values(by="StudyInstanceUID")
    unmelted_answers = unmelted_answers.sort_values(by="StudyInstanceUID")

    # extract the target columns
    y_true = unmelted_answers[column_order].to_numpy()
    y_pred = unmelted_submission[column_order].to_numpy()

    return y_true, y_pred


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    y_true, y_pred = prepare_for_metric(submission, answers)
    return custom_comp_weighted_loss(y_true, y_pred)



from typing import Dict
from dsat.prompts.common import create_draft_prompt

def create_stepwise_code_prompt(goal: str, plan: str, history: str, current_step: str, io_instructions: str) -> str:
    """Creates a prompt for generating code for a single step in a complex plan."""
    return (
        f"You are executing a complex data science plan step-by-step.\n\n"
        f"# OVERALL GOAL\n{goal}\n\n"
        f"# CRITICAL I/O REQUIREMENTS (MUST BE FOLLOWED)\n{io_instructions}\n\n"
        f"# OVERALL PLAN\n{plan}\n\n"
        f"# PREVIOUSLY EXECUTED STEPS (CODE & OUTPUT)\n{history}\n\n"
        f"# CURRENT STEP TO IMPLEMENT\n{current_step}\n\n"
        "Your task is to write the Python code ONLY for the CURRENT STEP. The code will be executed in a Jupyter-like environment, so you can assume variables from previous steps are in memory. Ensure your code adheres to the I/O requirements."
    )


def create_stepwise_debug_prompt(goal: str, plan: str, history: str, current_step: str, failed_code: str, error_history: str, io_instructions: str) -> str:
    """Creates a prompt for debugging a failed step in a stepwise execution."""
    return (
        f"You are debugging a complex data science plan being executed step-by-step.\n\n"
        f"# OVERALL GOAL\n{goal}\n\n"
        f"# CRITICAL I/O REQUIREMENTS (MUST BE FOLLOWED)\n{io_instructions}\n\n"
        f"# OVERALL PLAN\n{plan}\n\n"
        f"# PREVIOUSLY EXECUTED STEPS (CODE & OUTPUT)\n{history}\n\n"
        f"# FAILED STEP INSTRUCTION\n{current_step}\n\n"
        f"# MOST RECENT FAILED CODE\n```python\n{failed_code}\n```\n\n"
        f"# HISTORY OF FAILED ATTEMPTS FOR THIS STEP\n{error_history}\n\n"
        "Your task is to analyze the full history of errors for this step and provide the corrected Python code ONLY for the FAILED STEP. Assume variables from previous steps are still in memory. Ensure your corrected code adheres to the I/O requirements."
    )



import numpy as np
import pandas as pd
from sklearn import metrics

from mlebench.grade_helpers import InvalidSubmissionError


def insert_thresholds(fpr: np.ndarray, tpr: np.ndarray, tpr_thresholds: list) -> tuple:
    """
    Insert tpr_thresholds into the TPR and FPR arrays to ensure that the thresholds are present
    in the TPR array and the corresponding FPR values are interpolated.

    e.g.
    > tpr = [0.0, 0.25, 0.5, 0.75, 1.0]
    > fpr = [0.0, 0.1, 0.2, 0.3, 0.4]

    > tpr_thresholds = [0.0, 0.3, 0.4, 1.0]
    > fpr, tpr = insert_thresholds(fpr, tpr, tpr_thresholds)

    > print(tpr)
    > print(fpr)
    [0.0, 0.25, 0.3, 0.4, 0.5, 0.75, 1.0]
    [0.0, 0.1, 0.12, 0.16, 0.2, 0.3, 0.4]
    """
    fpr_ = fpr.tolist().copy()  # Don't modify the input arrays
    tpr_ = tpr.tolist().copy()
    for threshold in tpr_thresholds:
        if threshold not in tpr_:
            # Find the right position within tpr to insert the threshold
            for i, tpr_val in enumerate(tpr_):
                if tpr_val > threshold:
                    # Linear interpolation of fpr
                    new_fpr = fpr_[i - 1] + (fpr_[i] - fpr_[i - 1]) * (threshold - tpr_[i - 1]) / (
                        tpr_[i] - tpr_[i - 1]
                    )

                    tpr_.insert(i, threshold)
                    fpr_.insert(i, new_fpr)
                    break
    return np.array(fpr_), np.array(tpr_)


def alaska_weighted_auc(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Instead of the standard AUC, the competition uses a weighted AUC where different regions of
    the ROC curve are weighted differently.

    We compute the area under the curve segment by segment (HORIZONTAL segments between TPR pairs
    on the y-axis), and then compute a weighted average of the segments.

    For more details, see:
    www.kaggle.com/competitions/alaska2-image-steganalysis/overview/evaluation

    This particular implementation is adapted from:
    https://www.kaggle.com/code/anokas/weighted-auc-metric-updated
    (Key change vs the above implementation: The above implementation uses the `tpr` and `fpr`
    arrays from metrics.roc_curve as is, neglecting to handle the case where the `tpr` values
    don't line up nicely with the thresholds - leading to a situation where some segments either
    belong partially to the wrong threshold, or get skipped entirely. Our implementation fixes
    this by inserting the thresholds into the `tpr` and `fpr` arrays before computing the AUC.)
    """
    tpr_thresholds = [0.0, 0.4, 1.0]
    weights = [2, 1]

    fpr, tpr, thresholds = metrics.roc_curve(y_true, y_pred, pos_label=1)
    fpr, tpr = insert_thresholds(fpr, tpr, tpr_thresholds)

    # size of subsets
    areas = np.array(tpr_thresholds[1:]) - np.array(tpr_thresholds[:-1])

    competition_metric = 0
    # Compute AUC segment by segment (where each segment is a horizontal slice of the ROC curve
    # between a pair of consecutive TPR thresholds)
    for idx, weight in enumerate(weights):
        y_min = tpr_thresholds[idx]
        y_max = tpr_thresholds[idx + 1]

        # Here, we're creating new x and y arrays to calculate the AUC for this segment:

        # The segment arrays consist of the FPR and TPR values in the segment,
        mask = (y_min <= tpr) & (tpr <= y_max)
        if mask.sum() == 0:
            continue
        xs = fpr[mask]
        ys = tpr[mask]

        # plus a new point [1, y_max] which just closes the shape of this segment (draws a
        # horizontal line from the highest point in this segment to the right of the x-axis)
        xs = np.concatenate([xs, [1]])
        ys = np.concatenate([ys, [y_max]])
        ys = ys - y_min  # normalize such that curve starts at y=0
        score = metrics.auc(xs, ys)
        submetric = score * weight
        competition_metric += submetric

    # The total area is normalized by the sum of weights such that the final weighted AUC is between 0 and 1.
    normalization = np.dot(areas, weights)

    return competition_metric / normalization


def prepare_for_metric(submission: pd.DataFrame, answers: pd.DataFrame):
    if len(submission) != len(answers):
        raise InvalidSubmissionError("Submission and answers must have the same length")

    if "Id" not in submission.columns:
        raise InvalidSubmissionError("Submission must have an 'Id' column")

    if "Label" not in submission.columns:
        raise InvalidSubmissionError("Submission must have a 'Label' column")

    submission = submission.sort_values("Id")
    answers = answers.sort_values("Id")

    if (submission["Id"].values != answers["Id"].values).any():
        raise InvalidSubmissionError("Submission and answers must have the same ids")

    # Answers and submission must be numbers
    try:
        submission["Label"] = submission["Label"].astype(float)
    except ValueError:
        raise InvalidSubmissionError("Labels in submission must be numbers")
    answers["Label"] = answers["Label"].astype(float)

    # Cannot contain NaNs
    assert not answers["Label"].isnull().any(), "Answers cannot contain NaNs"
    if submission["Label"].isnull().any():
        raise InvalidSubmissionError("Submission cannot contain NaNs")

    return answers["Label"], submission["Label"]


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    y_true, y_pred = prepare_for_metric(submission, answers)

    return alaska_weighted_auc(np.array(y_true), np.array(y_pred))
